"""Interactive curvature analysis."""
from arpes.analysis.derivative import curvature, d1_along_axis, d2_along_axis
from arpes.analysis.filters import boxcar_filter, gaussian_filter
import numpy as np

from arpes.plotting.interactive_utils import BokehInteractiveTool
from arpes.utilities.funcutils import Debounce

__all__ = ["CurvatureTool"]


class CurvatureTool(BokehInteractiveTool):
    """An interactive Bokeh utility for selecting curvature analysis parameters."""

    # do not remove nans because they are used for masking in taking derivatives and smoothing
    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        """Loads image plot sizes from user settings."""
        super().__init__()

        self.load_settings(**kwargs)
        self.app_main_size = self.settings.get("app_main_size", 400)

    def tool_handler(self, doc):
        """Configures the Bokeh widgets for the curvature tool."""
        from bokeh.layouts import row, column
        from bokeh.models import widgets, Spacer
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.plotting import figure

        default_palette = self.default_palette

        x_coords, y_coords = self.arr.coords[self.arr.dims[1]], self.arr.coords[self.arr.dims[0]]
        self.app_context.update(
            {
                "data": self.arr,
                "cached_data": {},
                "gamma_cached_data": {},
                "plots": {},
                "data_range": self.arr.G.range(),
                "figures": {},
                "widgets": {},
                "color_maps": {},
            }
        )

        self.app_context["color_maps"]["d2"] = LinearColorMapper(
            default_palette,
            low=np.min(self.arr.values),
            high=np.max(self.arr.values),
            nan_color="black",
        )

        self.app_context["color_maps"]["curvature"] = LinearColorMapper(
            default_palette,
            low=np.min(self.arr.values),
            high=np.max(self.arr.values),
            nan_color="black",
        )

        self.app_context["color_maps"]["raw"] = LinearColorMapper(
            default_palette,
            low=np.min(self.arr.values),
            high=np.max(self.arr.values),
            nan_color="black",
        )

        plots, figures, data_range, cached_data, gamma_cached_data = (
            self.app_context["plots"],
            self.app_context["figures"],
            self.app_context["data_range"],
            self.app_context["cached_data"],
            self.app_context["gamma_cached_data"],
        )

        cached_data["raw"] = self.arr.values
        gamma_cached_data["raw"] = self.arr.values

        figure_kwargs = {
            "tools": ["reset", "wheel_zoom"],
            "plot_width": self.app_main_size,
            "plot_height": self.app_main_size,
            "min_border": 10,
            "toolbar_location": "left",
            "x_range": data_range["x"],
            "y_range": data_range["y"],
            "x_axis_location": "below",
            "y_axis_location": "right",
        }
        figures["d2"] = figure(title="d2 Spectrum", **figure_kwargs)

        figure_kwargs.update(
            {
                "y_range": self.app_context["figures"]["d2"].y_range,
                "x_range": self.app_context["figures"]["d2"].x_range,
                "toolbar_location": None,
                "y_axis_location": "left",
            }
        )

        figures["curvature"] = figure(title="Curvature", **figure_kwargs)
        figures["raw"] = figure(title="Raw Image", **figure_kwargs)

        figures["curvature"].yaxis.major_label_text_font_size = "0pt"

        # TODO add support for color mapper
        plots["d2"] = figures["d2"].image(
            [self.arr.values],
            x=data_range["x"][0],
            y=data_range["y"][0],
            dw=data_range["x"][1] - data_range["x"][0],
            dh=data_range["y"][1] - data_range["y"][0],
            color_mapper=self.app_context["color_maps"]["d2"],
        )
        plots["curvature"] = figures["curvature"].image(
            [self.arr.values],
            x=data_range["x"][0],
            y=data_range["y"][0],
            dw=data_range["x"][1] - data_range["x"][0],
            dh=data_range["y"][1] - data_range["y"][0],
            color_mapper=self.app_context["color_maps"]["curvature"],
        )
        plots["raw"] = figures["raw"].image(
            [self.arr.values],
            x=data_range["x"][0],
            y=data_range["y"][0],
            dw=data_range["x"][1] - data_range["x"][0],
            dh=data_range["y"][1] - data_range["y"][0],
            color_mapper=self.app_context["color_maps"]["raw"],
        )

        smoothing_sliders_by_name = {}
        smoothing_sliders = []  # need one for each axis
        axis_resolution = self.arr.G.stride(generic_dim_names=False)
        for dim in self.arr.dims:
            coords = self.arr.coords[dim]
            resolution = float(axis_resolution[dim])
            high_resolution = len(coords) / 3 * resolution
            low_resolution = resolution

            # could make this axis dependent for more reasonable defaults
            default = 15 * resolution

            if default > high_resolution:
                default = (high_resolution + low_resolution) / 2

            new_slider = widgets.Slider(
                title="{} Window".format(dim),
                start=low_resolution,
                end=high_resolution,
                step=resolution,
                value=default,
            )
            smoothing_sliders.append(new_slider)
            smoothing_sliders_by_name[dim] = new_slider

        n_smoothing_steps_slider = widgets.Slider(
            title="Smoothing Steps", start=0, end=5, step=1, value=2
        )
        beta_slider = widgets.Slider(title="β", start=-8, end=8, step=1, value=0)
        direction_select = widgets.Select(
            options=list(self.arr.dims),
            value="eV" if "eV" in self.arr.dims else self.arr.dims[0],  # preference to energy,
            title="Derivative Direction",
        )
        interleave_smoothing_toggle = widgets.Toggle(
            label="Interleave smoothing with d/dx", active=True, button_type="primary"
        )
        clamp_spectrum_toggle = widgets.Toggle(
            label="Clamp positive values to 0", active=True, button_type="primary"
        )
        filter_select = widgets.Select(
            options=["Gaussian", "Boxcar"], value="Boxcar", title="Type of Filter"
        )

        color_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Clip",
        )
        gamma_slider = widgets.Slider(start=0.1, end=4, value=1, step=0.1, title="Gamma")

        # don't need any cacheing here for now, might if this ends up being too slow
        def smoothing_fn(n_passes):
            if n_passes == 0:
                return lambda x: x

            filter_factory = {
                "Gaussian": gaussian_filter,
                "Boxcar": boxcar_filter,
            }.get(filter_select.value, boxcar_filter)

            filter_size = {d: smoothing_sliders_by_name[d].value for d in self.arr.dims}
            return filter_factory(filter_size, n_passes)

        @Debounce(0.25)
        def force_update():
            n_smoothing_steps = n_smoothing_steps_slider.value
            d2_data = self.arr
            if interleave_smoothing_toggle.active:
                f = smoothing_fn(n_smoothing_steps // 2)
                d2_data = d1_along_axis(f(d2_data), direction_select.value)
                f = smoothing_fn(n_smoothing_steps - (n_smoothing_steps // 2))
                d2_data = d1_along_axis(f(d2_data), direction_select.value)

            else:
                f = smoothing_fn(n_smoothing_steps)
                d2_data = d2_along_axis(f(self.arr), direction_select.value)

            d2_data.values[
                d2_data.values != d2_data.values
            ] = 0  # remove NaN values until Bokeh fixes NaNs over the wire
            if clamp_spectrum_toggle.active:
                d2_data.values = -d2_data.values
                d2_data.values[d2_data.values < 0] = 0
            cached_data["d2"] = d2_data.values
            gamma_cached_data["d2"] = d2_data.values ** gamma_slider.value
            plots["d2"].data_source.data = {"image": [gamma_cached_data["d2"]]}

            curv_smoothing_fn = smoothing_fn(n_smoothing_steps)
            smoothed_curvature_data = curv_smoothing_fn(self.arr)
            curvature_data = curvature(
                smoothed_curvature_data, self.arr.dims, beta=beta_slider.value
            )
            curvature_data.values[curvature_data.values != curvature_data.values] = 0
            if clamp_spectrum_toggle.active:
                curvature_data.values = -curvature_data.values
                curvature_data.values[curvature_data.values < 0] = 0

            cached_data["curvature"] = curvature_data.values
            gamma_cached_data["curvature"] = curvature_data.values ** gamma_slider.value
            plots["curvature"].data_source.data = {"image": [gamma_cached_data["curvature"]]}
            update_color_slider(color_slider.value)

        # TODO better integrate these, they can share code with the above if we are more careful.
        def take_d2(d2_data):
            n_smoothing_steps = n_smoothing_steps_slider.value
            if interleave_smoothing_toggle.active:
                f = smoothing_fn(n_smoothing_steps // 2)
                d2_data = d1_along_axis(f(d2_data), direction_select.value)
                f = smoothing_fn(n_smoothing_steps - (n_smoothing_steps // 2))
                d2_data = d1_along_axis(f(d2_data), direction_select.value)

            else:
                f = smoothing_fn(n_smoothing_steps)
                d2_data = d2_along_axis(f(self.arr), direction_select.value)

            d2_data.values[
                d2_data.values != d2_data.values
            ] = 0  # remove NaN values until Bokeh fixes NaNs over the wire
            if clamp_spectrum_toggle.active:
                d2_data.values = -d2_data.values
                d2_data.values[d2_data.values < 0] = 0

            return d2_data

        def take_curvature(curvature_data, curve_dims):
            curv_smoothing_fn = smoothing_fn(n_smoothing_steps_slider.value)
            smoothed_curvature_data = curv_smoothing_fn(curvature_data)
            curvature_data = curvature(smoothed_curvature_data, curve_dims, beta=beta_slider.value)
            curvature_data.values[curvature_data.values != curvature_data.values] = 0
            if clamp_spectrum_toggle.active:
                curvature_data.values = -curvature_data.values
                curvature_data.values[curvature_data.values < 0] = 0

            return curvature_data

        # These functions will always be linked to the current context of the curvature tool.
        self.app_context["d2_fn"] = take_d2
        self.app_context["curvature_fn"] = take_curvature

        def force_update_change_wrapper(attr, old, new):
            if old != new:
                force_update()

        def force_update_click_wrapper(event):
            force_update()

        @Debounce(0.1)
        def update_color_slider(new):
            def update_plot(name, data):
                low, high = np.min(data), np.max(data)
                dynamic_range = high - low
                self.app_context["color_maps"][name].update(
                    low=low + new[0] / 100 * dynamic_range, high=low + new[1] / 100 * dynamic_range
                )

            update_plot("d2", gamma_cached_data["d2"])
            update_plot("curvature", gamma_cached_data["curvature"])
            update_plot("raw", gamma_cached_data["raw"])

        @Debounce(0.1)
        def update_gamma_slider(new):
            gamma_cached_data["d2"] = cached_data["d2"] ** new
            gamma_cached_data["curvature"] = cached_data["curvature"] ** new
            gamma_cached_data["raw"] = cached_data["raw"] ** new
            update_color_slider(color_slider.value)

        def update_color_handler(attr, old, new):
            update_color_slider(new)

        def update_gamma_handler(attr, old, new):
            update_gamma_slider(new)

        layout = column(
            row(
                column(
                    self.app_context["figures"]["d2"], interleave_smoothing_toggle, direction_select
                ),
                column(
                    self.app_context["figures"]["curvature"], beta_slider, clamp_spectrum_toggle
                ),
                column(self.app_context["figures"]["raw"], color_slider, gamma_slider),
            ),
            column(
                filter_select,
                *smoothing_sliders,
                n_smoothing_steps_slider,
            ),
            Spacer(height=100),
        )

        # Attach event handlers
        for w in (
            n_smoothing_steps_slider,
            beta_slider,
            direction_select,
            *smoothing_sliders,
            filter_select,
        ):
            w.on_change("value", force_update_change_wrapper)

        interleave_smoothing_toggle.on_click(force_update_click_wrapper)
        clamp_spectrum_toggle.on_click(force_update_click_wrapper)

        color_slider.on_change("value", update_color_handler)
        gamma_slider.on_change("value", update_gamma_handler)

        force_update()

        doc.add_root(layout)
        doc.title = "Curvature Tool"



"""Curve fitting models with two independent variables."""

from arpes.constants import HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ
from lmfit.models import update_param_vals
import numpy as np
import lmfit as lf

from .x_model_mixin import XModelMixin

__all__ = [
    "Gaussian2DModel",
    "EffectiveMassModel",
]

any_dim_sentinel = None


class Gaussian2DModel(XModelMixin):
    """2D Gaussian fitting."""

    n_dims = 2
    dimension_order = [any_dim_sentinel, any_dim_sentinel]

    @staticmethod
    def gaussian_2d_bkg(
        x, y, amplitude=1, xc=0, yc=0, sigma_x=0.1, sigma_y=0.1, const_bkg=0, x_bkg=0, y_bkg=0
    ):
        """Defines a multidimensional axis aligned normal."""
        bkg = np.outer(x * 0 + 1, y_bkg * y) + np.outer(x * x_bkg, y * 0 + 1) + const_bkg
        # make the 2D Gaussian matrix
        gauss = (
            amplitude
            * np.exp(
                -(
                    (x[:, None] - xc) ** 2 / (2 * sigma_x ** 2)
                    + (y[None, :] - yc) ** 2 / (2 * sigma_y ** 2)
                )
            )
            / (2 * np.pi * sigma_x * sigma_y)
        )

        # flatten the 2D Gaussian down to 1D
        return np.ravel(gauss + bkg)

    def __init__(
        self, independent_vars=("x", "y"), prefix="", missing="raise", name=None, **kwargs
    ):
        """Sets reasonable constraints on the width and constraints the amplitude to be positive."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.gaussian_2d_bkg, **kwargs)

        self.set_param_hint("sigma_x", min=0.0)
        self.set_param_hint("sigma_y", min=0.0)
        self.set_param_hint("amplitude", min=0.0)


class EffectiveMassModel(XModelMixin):
    """A two dimensional model for a quadratic distribution of Lorentzians.

    This can be used for "global" fitting of the effective mass of a band, providing higher quality
    results than iterative fitting for the lineshapes and then performing a
    quadratic fit to the centers.

    This model also provides a representative example of how to implement (2+)D "global"
    lmfit.Model classes which are xarray coordinate aware.

        `dimension_order`: allows specifying allowed sets of dimension for the global fit (here eV
        must be one axis, while either kp or phi is allowed for the other.
    """

    n_dims = 2
    dimension_order = ["eV", ["kp", "phi"]]

    @staticmethod
    def effective_mass_bkg(
        eV,
        kp,
        m_star=0,
        k_center=0,
        eV_center=0,
        gamma=1,
        amplitude=1,
        amplitude_k=0,
        const_bkg=0,
        k_bkg=0,
        eV_bkg=0,
    ):
        """Model implementation function for simultaneous 2D curve fitting of band effective mass.

        Allows for an affine background in each dimension, together with variance in the band intensity
        along the band, as a very simple model of matrix elements. Together with prenormalizing your data
        this should allow reasonable fits of a lot of typical ARPES data.
        """
        bkg = np.outer(eV * 0 + 1, k_bkg * kp) + np.outer(eV_bkg * eV, kp * 0 + 1) + const_bkg

        # check units
        dk = kp - k_center
        offset = HBAR_SQ_EV_PER_ELECTRON_MASS_ANGSTROM_SQ * dk ** 2 / (2 * m_star + 1e-6)
        eVk = np.outer(eV, kp * 0 + 1)
        coherent = (
            (amplitude + amplitude_k * dk)
            * (1 / (2 * np.pi))
            * gamma
            / ((eVk - eV_center + offset) ** 2 + (0.5 * gamma) ** 2)
        )
        return (coherent + bkg).ravel()

    def __init__(
        self, independent_vars=("eV", "kp"), prefix="", missing="raise", name=None, **kwargs
    ):
        """Mostly just set parameter hints to physically realistic values here."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.effective_mass_bkg, **kwargs)

        self.set_param_hint("gamma", min=0.0)
        self.set_param_hint("amplitude", min=0.0)

    def guess(self, data, eV=None, kp=None, phi=None, **kwargs):
        """Use heuristics to estimate the model parameters."""
        momentum = kp if kp is not None else phi
        try:
            momentum = momentum.values
            eV = eV.values
            data = data.values
        except AttributeError:
            pass

        pars = self.make_params()

        pars["%sm_star" % self.prefix].set(value=1)

        pars["%sk_center" % self.prefix].set(value=np.mean(momentum))
        pars["%seV_center" % self.prefix].set(value=np.mean(eV))

        pars["%samplitude" % self.prefix].set(value=np.mean(np.mean(data, axis=0)))
        pars["%sgamma" % self.prefix].set(value=0.25)

        pars["%samplitude_k" % self.prefix].set(value=0)  # can definitely improve here

        # Crude estimate of the background
        left, right = np.mean(data[:5, :], axis=0), np.mean(data[-5:, :], axis=0)
        top, bottom = np.mean(data[:, :5], axis=0), np.mean(data[:, -5:], axis=0)
        left, right = np.percentile(left, 10), np.percentile(right, 10)
        top, bottom = np.percentile(top, 10), np.percentile(bottom, 10)

        pars["%sconst_bkg" % self.prefix].set(value=np.min(np.array([left, right, top, bottom])))
        pars["%sk_bkg" % self.prefix].set(value=(bottom - top) / (eV[-1] - eV[0]))
        pars["%seV_bkg" % self.prefix].set(value=(right - left) / (momentum[-1] - momentum[0]))

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC



"""Some miscellaneous model definitions."""

from lmfit.models import update_param_vals
import lmfit as lf
import numpy as np

from .x_model_mixin import XModelMixin

__all__ = [
    "QuadraticModel",
    "FermiVelocityRenormalizationModel",
    "LogRenormalizationModel",
]


class QuadraticModel(XModelMixin):
    """A model for fitting a quadratic function."""

    @staticmethod
    def quadratic(x, a=1, b=0, c=0):
        """Quadratic polynomial."""
        return a * x ** 2 + b * x + c

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Just defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.quadratic, **kwargs)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for parameter guesses."""
        pars = self.make_params()

        pars["%sa" % self.prefix].set(value=0)
        pars["%sb" % self.prefix].set(value=0)
        pars["%sc" % self.prefix].set(value=data.mean())

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class FermiVelocityRenormalizationModel(XModelMixin):
    """A model for Logarithmic Renormalization to Fermi Velocity in Dirac Materials."""

    @staticmethod
    def fermi_velocity_renormalization_mfl(x, n0, v0, alpha, eps):
        """A model for Logarithmic Renormalization to Fermi Velocity in Dirac Materials.

        Args:
            x: value to evaluate fit at (carrier density)
            n0: Value of carrier density at cutoff energy for validity of Dirac fermions
            v0: Bare velocity
            alpha: Fine structure constant
            eps: Graphene Dielectric constant
        """
        #     y = v0 * (rs/np.pi)*(5/3 + np.log(rs))+(rs/4)*np.log(kc/np.abs(kF))
        fx = v0 * (1 + (alpha / (1 + eps)) * np.log(n0 / np.abs(x)))
        fx2 = v0 * (1 + (alpha / (1 + eps * np.abs(x))) * np.log(n0 / np.abs(x)))
        fx3 = v0 * (1 + (alpha / (1 + eps * x ** 2)) * np.log(n0 / np.abs(x)))
        # return v0 + v0*(alpha/(8*eps))*np.log(n0/x)
        return fx3

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Sets physically reasonable constraints on parameter values."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.fermi_velocity_renormalization_mfl, **kwargs)

        self.set_param_hint("alpha", min=0.0)
        self.set_param_hint("n0", min=0.0)
        self.set_param_hint("eps", min=0.0)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for parameter estimation."""
        pars = self.make_params()

        # pars['%sn0' % self.prefix].set(value=10)
        # pars['%seps' % self.prefix].set(value=8)
        # pars['%svF' % self.prefix].set(value=(data.max()-data.min())/(kC-kD))

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC


class LogRenormalizationModel(XModelMixin):
    """A model for Logarithmic Renormalization to Linear Dispersion in Dirac Materials."""

    @staticmethod
    def log_renormalization(x, kF=1.6, kD=1.6, kC=1.7, alpha=0.4, vF=1e6):
        """Logarithmic correction to linear dispersion near charge neutrality in Dirac materials.

        As examples, this can be used to study the low energy physics in high quality ARPES spectra of graphene
        or topological Dirac semimetals.

        Args:
            x: The coorindates for the fit
            k: value to evaluate fit at
            kF: Fermi wavevector
            kD: Dirac point
            alpha: Fine structure constant
            vF: Bare Band Fermi Velocity
            kC: Cutoff Momentum
        """
        dk = x - kF
        dkD = x - kD
        return -vF * np.abs(dkD) + (alpha / 4) * vF * dk * np.log(np.abs(kC / dkD))

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """The fine structure constant and velocity must be nonnegative, so we will constrain them here."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.log_renormalization, **kwargs)

        self.set_param_hint("alpha", min=0.0)
        self.set_param_hint("vF", min=0.0)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for actually making parameter estimates here."""
        pars = self.make_params()

        pars["%skC" % self.prefix].set(value=1.7)

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC



"""Plotting routines for making the classic stacked line plots.

Think the album art for "Unknown Pleasures".
"""

import matplotlib.colors
import matplotlib.pyplot as plt
import numpy as np
from math import ceil
from matplotlib import cm
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

import xarray as xr
from arpes.analysis.general import rebin
from arpes.plotting.utils import (
    colorbarmaps_for_axis,
    generic_colorbarmap_for_data,
    fancy_labels,
    path_for_plot,
    label_for_dim,
)
from arpes.provenance import save_plot_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum, lift_spectrum

__all__ = (
    "stack_dispersion_plot",
    "flat_stack_plot",
    "overlapped_stack_dispersion_plot",
)


@save_plot_provenance
def flat_stack_plot(
    data: DataType,
    stack_axis=None,
    fermi_level=True,
    cbarmap=None,
    ax=None,
    mode="line",
    title=None,
    out=None,
    transpose=False,
    **kwargs,
):
    """Generates a stack plot with all the lines distinguished by color rather than offset."""
    data = normalize_to_spectrum(data)
    if len(data.dims) != 2:
        raise ValueError(
            "In order to produce a stack plot, data must be image-like."
            "Passed data included dimensions: {}".format(data.dims)
        )

    fig = None
    inset_ax = None
    if ax is None:
        fig, ax = plt.subplots(
            figsize=kwargs.get(
                "figsize",
                (
                    7,
                    5,
                ),
            )
        )
        inset_ax = inset_axes(ax, width="40%", height="5%", loc=1)

    if stack_axis is None:
        stack_axis = data.dims[0]

    skip_colorbar = True
    if cbarmap is None:
        skip_colorbar = False
        try:
            cbarmap = colorbarmaps_for_axis[stack_axis]
        except KeyError:
            cbarmap = generic_colorbarmap_for_data(
                data.coords[stack_axis], ax=inset_ax, ticks=kwargs.get("ticks")
            )

    cbar, cmap = cbarmap

    # should be exactly two
    other_dim = [d for d in data.dims if d != stack_axis][0]
    other_coord = data.coords[other_dim]

    if not isinstance(cmap, matplotlib.colors.Colormap):
        # do our best
        try:
            cmap = cmap()
        except:
            # might still be fine
            pass

    if "eV" in data.dims and "eV" != stack_axis and fermi_level:
        if transpose:
            ax.axhline(0, color="red", alpha=0.8, linestyle="--", linewidth=1)
        else:
            ax.axvline(0, color="red", alpha=0.8, linestyle="--", linewidth=1)

    # meat of the plotting
    for coord_dict, marginal in list(data.G.iterate_axis(stack_axis)):
        if transpose:
            if mode == "line":
                ax.plot(
                    marginal.values,
                    marginal.coords[marginal.dims[0]].values,
                    color=cmap(coord_dict[stack_axis]),
                    **kwargs,
                )
            else:
                assert mode == "scatter"
                raise NotImplementedError
        else:
            if mode == "line":
                marginal.plot(ax=ax, color=cmap(coord_dict[stack_axis]), **kwargs)
            else:
                assert mode == "scatter"
                ax.scatter(
                    *marginal.G.to_arrays(),
                    color=cmap(coord_dict[stack_axis]),
                    **kwargs,
                )
                ax.set_xlabel(marginal.dims[0])

    ax.set_xlabel(label_for_dim(data, ax.get_xlabel()))
    ax.set_ylabel("Spectrum Intensity (arb).")
    ax.set_title(title, fontsize=14)
    ax.set_xlim([other_coord.min().item(), other_coord.max().item()])

    try:
        if inset_ax is not None and not skip_colorbar:
            inset_ax.set_xlabel(stack_axis, fontsize=16)
            fancy_labels(inset_ax)

            cbar(ax=inset_ax, **kwargs)
    except TypeError:
        # already rendered
        pass

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax


@lift_spectrum
def stack_dispersion_plot(
    spectrum: xr.DataArray,
    stack_axis: str = None,
    data_scaling: float = 2.0,
    max_stacks: int = 15,
    ax: plt.Axes | None = None,
    **kwargs,
) -> tuple[plt.Figure, plt.Axes]:
    """Generate a stack plot of a 2D spectrum along the specified `stack_axis`.

    Each cut is scaled to fit within the average cut offset multiplied by
    `data_scaling`. Cuts are offset so they intersect the y-axis near their cut value.

    Args
    ----
        spectrum : DataArray
            dataset to plot
        stack_axis : str
            axis to stack along (defaults to the zeroth axis)
        data_scaling : float
            scaling factor for the data
        max_stacks : int
            maximum number of stacks to plot

    Returns
    -------
        fig, ax: figure and axis of the plot
    """
    axes = list(spectrum.dims)
    if len(axes) != 2:
        raise ValueError(f"Spectrum must be 2D, but has dimensions {spectrum.dims}")
    stack_axis = axes[0] if stack_axis is None else stack_axis
    try:
        axes.remove(stack_axis)
    except ValueError:
        raise ValueError(
            f"stack_axis, {stack_axis}, is not one of the data's axes: {axes}"
        )
    plot_axis = axes[0]

    spectrum = spectrum.sortby(stack_axis)
    if max_stacks < len(spectrum.coords[stack_axis]):
        stack_data = rebin(
            spectrum,
            reduction={stack_axis: ceil(len(spectrum[stack_axis]) / max_stacks)},
        )
    else:
        stack_data = spectrum.copy(deep=True)

    cut_values = stack_data.coords[stack_axis].values
    x = stack_data.coords[plot_axis].values

    def normalize_by_max(y: np.ndarray, max: float):
        return (y - y.min()) / max

    raw_ys: list[np.ndarray] = [
        y_values for y_values in stack_data.transpose(stack_axis, ...).values
    ]
    max_height = max([y.max() - y.min() for y in raw_ys])
    normalized_ys = [normalize_by_max(y, max_height) for y in raw_ys]

    average_cut_offset = np.mean(np.diff(cut_values))
    average_y_intercept = np.mean([y[0] for y in normalized_ys])
    scaled_ys = np.array(
        [
            ((y - average_y_intercept) * data_scaling * average_cut_offset) + cut_value
            for y, cut_value in zip(normalized_ys, cut_values)
        ]
    )

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = ax.get_figure()
    ax.plot(x, scaled_ys.transpose(), **kwargs)

    ax.set_xlim(x.min(), x.max())
    ax.set_xlabel(plot_axis)
    ax.set_ylabel(stack_axis)

    return fig, ax


@save_plot_provenance
def overlapped_stack_dispersion_plot(
    data: DataType,
    stack_axis=None,
    ax=None,
    title=None,
    out=None,
    max_stacks=100,
    use_constant_correction=False,
    transpose=False,
    negate=False,
    s=1,
    scale_factor=None,
    linewidth=1,
    palette=None,
    **kwargs,
):
    data = normalize_to_spectrum(data)

    if stack_axis is None:
        stack_axis = data.dims[0]

    other_axes = list(data.dims)
    other_axes.remove(stack_axis)
    other_axis = other_axes[0]

    stack_coord = data.coords[stack_axis]
    if len(stack_coord.values) > max_stacks:
        data = rebin(
            data,
            reduction=dict(
                [[stack_axis, int(np.ceil(len(stack_coord.values) / max_stacks))]]
            ),
        )

    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=(7, 7))

    if title is None:
        title = "{} Stack".format(data.S.label.replace("_", " "))

    max_over_stacks = np.max(data.values)

    cvalues = data.coords[other_axis].values
    if scale_factor is None:
        maximum_deviation = -np.inf

        for _, marginal in data.G.iterate_axis(stack_axis):
            marginal_values = -marginal.values if negate else marginal.values
            marginal_offset, right_marginal_offset = (
                marginal_values[0],
                marginal_values[-1],
            )

            if use_constant_correction:
                true_ys = marginal_values - marginal_offset
            else:
                true_ys = marginal_values - np.linspace(
                    marginal_offset, right_marginal_offset, len(marginal_values)
                )

            maximum_deviation = np.max([maximum_deviation] + list(np.abs(true_ys)))

        scale_factor = 0.02 * (np.max(cvalues) - np.min(cvalues)) / maximum_deviation

    iteration_order = -1  # might need to fiddle with this in certain cases
    for coord_dict, marginal in list(data.G.iterate_axis(stack_axis))[
        ::iteration_order
    ]:
        coord_value = coord_dict[stack_axis]

        xs = cvalues
        marginal_values = -marginal.values if negate else marginal.values
        marginal_offset, right_marginal_offset = marginal_values[0], marginal_values[-1]

        if use_constant_correction:
            true_ys = (marginal_values - marginal_offset) / max_over_stacks
            ys = scale_factor * true_ys + coord_value
        else:
            true_ys = (
                marginal_values
                - np.linspace(
                    marginal_offset, right_marginal_offset, len(marginal_values)
                )
            ) / max_over_stacks
            ys = scale_factor * true_ys + coord_value

        raw_colors = "black"
        if palette:
            if isinstance(palette, str):
                palette = cm.get_cmap(palette)
            raw_colors = palette(np.abs(true_ys / max_over_stacks))

        if transpose:
            xs, ys = ys, xs

        if isinstance(raw_colors, str):
            plt.plot(xs, ys, linewidth=linewidth, color=raw_colors, **kwargs)
        else:
            plt.scatter(xs, ys, color=raw_colors, s=s, **kwargs)

    x_label = other_axis
    y_label = stack_axis

    if transpose:
        x_label, y_label = y_label, x_label

    ax.set_xlabel(label_for_dim(data, x_label))
    ax.set_ylabel(label_for_dim(data, y_label))

    ax.set_title(title)

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    plt.show()

    return fig, ax



"""A utility for selecting paths on a marginal of your data."""
import numpy as np

import xarray as xr
from arpes.analysis.path import select_along_path
from arpes.exceptions import AnalysisError
from arpes.plotting.interactive_utils import CursorTool, SaveableTool
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum

__all__ = ["path_tool"]


class PathTool(SaveableTool, CursorTool):
    """Tool to allow drawing paths on data, creating selections, and masking regions around paths.

    Integrates with the tools in arpes.analysis.path
    """

    auto_zero_nans = False
    auto_rebin = False

    def __init__(self, **kwargs):
        super().__init__(kwargs.pop("name", None))

        self.load_settings(**kwargs)

        self.app_main_size = self.settings.get("app_main_size", 600)
        self.app_marginal_size = self.settings.get("app_main_size", 300)
        self.pointer_mode = "path"

    def tool_handler(self, doc):
        from bokeh import events
        from bokeh.layouts import row, column
        from bokeh.models.mappers import LinearColorMapper
        from bokeh.models import widgets, warnings
        from bokeh.plotting import figure

        if len(self.arr.shape) != 2:
            raise AnalysisError("Cannot use path tool on non image-like spectra")

        arr = self.arr
        x_coords, y_coords = arr.coords[arr.dims[0]], arr.coords[arr.dims[1]]

        default_palette = self.default_palette

        self.app_context.update(
            {
                "path_options": [],
                "active_path": None,
                "paths": {},
                "data": arr,
                "data_range": {
                    "x": (np.min(x_coords.values), np.max(x_coords.values)),
                    "y": (np.min(y_coords.values), np.max(y_coords.values)),
                },
            }
        )

        self.cursor = [np.mean(self.data_range["x"]), np.mean(self.data_range["y"])]

        self.color_maps["main"] = LinearColorMapper(
            default_palette, low=np.min(arr.values), high=np.max(arr.values), nan_color="black"
        )

        main_tools = ["wheel_zoom", "tap", "reset"]
        main_title = "Path Tool: WARNING Unidentified"

        try:
            main_title = "Path Tool: {}".format(arr.S.label[:60])
        except:
            pass

        self.figures["main"] = figure(
            tools=main_tools,
            plot_width=self.app_main_size,
            plot_height=self.app_main_size,
            min_border=10,
            min_border_left=20,
            toolbar_location="left",
            x_axis_location="below",
            y_axis_location="right",
            title=main_title,
            x_range=self.data_range["x"],
            y_range=self.data_range["y"],
        )

        self.figures["main"].xaxis.axis_label = arr.dims[0]
        self.figures["main"].yaxis.axis_label = arr.dims[1]

        self.plots["main"] = self.figures["main"].image(
            [np.asarray(arr.values.T)],
            x=self.data_range["x"][0],
            y=self.data_range["y"][0],
            dw=self.data_range["x"][1] - self.data_range["x"][0],
            dh=self.data_range["y"][1] - self.data_range["y"][0],
            color_mapper=self.color_maps["main"],
        )

        self.plots["paths"] = self.figures["main"].multi_line(
            xs=[], ys=[], line_color="white", line_width=2
        )

        self.add_cursor_lines(self.figures["main"])

        def add_point_to_path():
            if self.active_path in self.paths:
                self.paths[self.active_path]["points"].append(list(self.cursor))
                update_path_display()

            self.save_app()

        def click_main_image(event):
            self.cursor = [event.x, event.y]
            if self.pointer_mode == "path":
                add_point_to_path()

        POINTER_MODES = [
            ("Cursor", "cursor"),
            ("Path", "path"),
        ]

        def convert_to_xarray():
            """Creates a Dataset consisting of one array for each path.

            For each of the paths, we will create a dataset which has an index dimension,
            and datavariables for each of the coordinate dimensions
            """

            def convert_single_path_to_xarray(points):
                vars = {d: np.array([p[i] for p in points]) for i, d in enumerate(self.arr.dims)}
                coords = {
                    "index": np.array(range(len(points))),
                }
                vars = {k: xr.DataArray(v, coords=coords, dims=["index"]) for k, v in vars.items()}
                return xr.Dataset(data_vars=vars, coords=coords)

            return {
                path["name"]: convert_single_path_to_xarray(path["points"])
                for path in self.paths.values()
            }

        def select(data=None, **kwargs):
            if data is None:
                data = self.arr

            if len(self.paths) > 1:
                warnings.warn("Only using first path.")

            return select_along_path(
                path=list(convert_to_xarray().items())[0][1], data=data, **kwargs
            )

        self.app_context["to_xarray"] = convert_to_xarray
        self.app_context["select"] = select

        pointer_dropdown = widgets.Dropdown(
            label="Pointer Mode", button_type="primary", menu=POINTER_MODES
        )
        self.path_dropdown = widgets.Dropdown(
            label="Active Path", button_type="primary", menu=self.path_options
        )

        path_name_input = widgets.TextInput(placeholder="Path name...")
        add_path_button = widgets.Button(label="Add Path")

        clear_path_button = widgets.Button(label="Clear Path")
        remove_path_button = widgets.Button(label="Remove Path")

        main_color_range_slider = widgets.RangeSlider(
            start=0,
            end=100,
            value=(
                0,
                100,
            ),
            title="Color Range",
        )

        def add_path(path_name):
            if path_name not in self.paths:
                self.path_options.append(
                    (
                        path_name,
                        path_name,
                    )
                )
                self.path_dropdown.menu = self.path_options
                self.paths[path_name] = {
                    "points": [],
                    "name": path_name,
                }

                if self.active_path is None:
                    self.active_path = path_name

                self.save_app()

        def on_change_active_path(event):
            path_id = event.item
            self.debug_text = path_id
            self.app_context["active_path"] = path_id
            self.active_path = path_id
            self.save_app()

        def on_change_pointer_mode(event):
            pointer_mode = event.item
            self.app_context["pointer_mode"] = pointer_mode
            self.pointer_mode = pointer_mode
            self.save_app()

        def update_path_display():
            self.plots["paths"].data_source.data = {
                "xs": [[point[0] for point in p["points"]] for p in self.paths.values()],
                "ys": [[point[1] for point in p["points"]] for p in self.paths.values()],
            }
            self.save_app()

        self.update_path_display = update_path_display

        def on_clear_path():
            if self.active_path in self.paths:
                self.paths[self.active_path]["points"] = []
                update_path_display()

        def on_remove_path():
            if self.active_path in self.paths:
                del self.paths[self.active_path]
                new_path_options = [b for b in self.path_options if b[0] != self.active_path]
                self.path_dropdown.menu = new_path_options
                self.path_options = new_path_options
                self.active_path = None
                update_path_display()

        # Attach callbacks
        main_color_range_slider.on_change("value", self.update_colormap_for("main"))

        self.figures["main"].on_event(events.Tap, click_main_image)
        self.path_dropdown.on_click(on_change_active_path)
        pointer_dropdown.on_click(on_change_pointer_mode)
        add_path_button.on_click(lambda: add_path(path_name_input.value))
        clear_path_button.on_click(on_clear_path)
        remove_path_button.on_click(on_remove_path)

        layout = row(
            column(self.figures["main"]),
            column(
                column(
                    pointer_dropdown,
                    self.path_dropdown,
                ),
                row(
                    path_name_input,
                    add_path_button,
                ),
                row(
                    clear_path_button,
                    remove_path_button,
                ),
                column(
                    self._cursor_info,
                    main_color_range_slider,
                ),
                self.debug_div,
            ),
        )

        doc.add_root(layout)
        doc.title = "Path Tool"
        self.load_app()
        self.save_app()

    def serialize(self):
        return {
            "active_path": self.active_path,
            "path_options": self.path_options,
            "paths": self.paths,
            "cursor": self.cursor,
        }

    def deserialize(self, json_data):
        self.cursor = json_data.get("cursor", [0, 0])

        self.app_context["paths"] = json_data.get("paths", {}) or {}
        self.app_context["path_options"] = json_data.get("path_options", [])

        self.path_dropdown.menu = self.app_context["path_options"]

        self.update_path_display()


def path_tool(data: DataType, **kwargs):
    """Opens the path tool for the given data."""
    data = normalize_to_spectrum(data)

    tool = PathTool(**kwargs)
    return tool.make_tool(data)



"""Infrastructure code for interactive Bokeh based analysis tools."""

import json
import os
import warnings
from abc import ABC, abstractmethod
from pathlib import Path

import numpy as np

from arpes.settings import SETTINGS
import colorcet as cc
import xarray as xr
from arpes.analysis.general import rebin
from arpes.io import load_data
from arpes.utilities import deep_equals
from typing import List, Union

__all__ = (
    "BokehInteractiveTool",
    "CursorTool",
)


class CursorTool:
    """A base class for a Bokeh based analysis application which using a cursor into a data volume."""

    _cursor = None
    _cursor_info = None
    _horiz_cursor_x = None
    _horiz_cursor_y = None
    _vert_cursor_x = None
    _vert_cursor_y = None
    _cursor_lines = None
    _cursor_dims = None

    def __init__(self):
        """Initializes application context and not much else."""
        self.data_range = {}
        self.arr = None
        self.app_context = {}

    @property
    def cursor_dims(self) -> List[str]:
        """The dimesnion names for the current cursor order."""
        return self._cursor_dims

    @cursor_dims.setter
    def cursor_dims(self, value):
        self._cursor_dims = value

    @property
    def cursor_dict(self):
        """The location of the cursor in the data volume, as a dim=value dict."""
        if self._cursor_dims is None:
            return None

        return dict(zip(self.cursor_dims, self.cursor))

    @property
    def cursor(self):
        """The location of the cursor in the data volume."""
        return self._cursor

    def add_cursor_lines(self, figure):
        """Adds the standard X and Y cursors to a figure."""
        cursor_lines = figure.multi_line(
            xs=[self._horiz_cursor_x, self._vert_cursor_x],
            ys=[self._horiz_cursor_y, self._vert_cursor_y],
            line_color="white",
            line_width=2,
            line_dash="dotted",
        )
        self._cursor_lines = cursor_lines
        return cursor_lines

    @cursor.setter
    def cursor(self, values):
        from bokeh.models.widgets.markups import Div

        if self._cursor_dims is None:
            try:
                self._cursor_dims = list(self.arr.dims)
            except AttributeError:
                pass

        self._cursor = values
        if self._cursor_info is None:
            self._cursor_info = Div(text="")
            self._horiz_cursor_x = list(self.data_range["x"])
            self._horiz_cursor_y = [0, 0]
            self._vert_cursor_x = [0, 0]
            self._vert_cursor_y = list(self.data_range["y"])
        else:
            self._vert_cursor_x[0] = self._vert_cursor_x[1] = self.cursor[0]
            self._horiz_cursor_y[0] = self._horiz_cursor_y[1] = self.cursor[1]

        self._cursor_info.text = "<h2>Cursor:</h2><span>({})</span>".format(
            ", ".join("{0:.3f}".format(c) for c in self.cursor)
        )

        if self._cursor_lines is not None:
            self._cursor_lines.data_source.data = {
                "xs": [self._horiz_cursor_x, self._vert_cursor_x],
                "ys": [self._horiz_cursor_y, self._vert_cursor_y],
            }

        try:
            self.app_context["cursor_dict"] = dict(zip(self._cursor_dims, self.cursor))
            # self.app_context['full_cursor'] =
        except AttributeError:
            pass


class BokehInteractiveTool(ABC):
    """Base class for Bokeh based analysis applications.

    You should view this as deprecated in light of newer variants based on Qt
    which are much much more performant. There is no way of doing array memory sharing
    with Jupyter so high performance browser based applications for PyARPES are probably
    not going to happen anytime soon.
    """

    auto_rebin = True
    auto_zero_nans = True
    rebin_size = 800

    _debug_div = None

    @property
    def debug_div(self):
        """A debug element which makes developing tools with Bokeh more straightforward."""
        from bokeh.models.widgets.markups import Div

        if self._debug_div is None:
            self._debug_div = Div(text="", width=300, height=100)

        return self._debug_div

    def __setattr__(self, name, value):
        """Overrides __setattr__ to handle setting debug info into a div so it's visible."""
        if name == "debug_text":
            self._debug_div.text = value

        super().__setattr__(name, value)

    def update_colormap_for(self, plot_name):
        """Sets the colormap on the plot `plot_name` to have an appropriate range."""

        def update_plot_colormap(attr, old, new):
            plot_data = self.plots[plot_name].data_source.data["image"]
            low, high = np.min(plot_data), np.max(plot_data)
            dynamic_range = high - low
            self.color_maps[plot_name].update(
                low=low + new[0] / 100 * dynamic_range,
                high=low + new[1] / 100 * dynamic_range,
            )

        return update_plot_colormap

    def init_bokeh_server(self):
        """Tells Bokeh to send output to Jupyter with a relatively long timeout.

        The long timeout is allows for slowish tool startup and transport of large arrays
        over HTTP to the running JS application.
        """
        from bokeh.io import output_notebook

        if "bokeh_configured" not in SETTINGS:
            SETTINGS["bokeh_configured"] = True

            # use a longer load_timeout for heavy tools
            output_notebook(hide_banner=True, load_timeout=10000)

            # Don't need to manually start a server in the manner of
            # https://matthewrocklin.com/blog//work/2017/06/28/simple-bokeh-server
            # according to
            # https://github.com/bokeh/bokeh/blob/0.12.10/examples/howto/server_embed/notebook_embed.ipynb

    def load_settings(self, **kwargs):
        """Loads a user's settings for interactive tools into the tool.

        Various settings, like the sizes of widgets and panels can be set in user
        settings overrides, and are read here.
        """
        self.settings = SETTINGS.get("interactive", {}).copy()
        for k, v in kwargs.items():
            if k not in self.settings:
                self.settings[k] = v

    @property
    def default_palette(self):
        """Resolves user settings for a color palette to an actual matplotlib palette."""
        from bokeh import palettes

        palette_options = {
            "viridis": palettes.viridis(256),
            "magma": palettes.magma(256),
            "coolwarm": cc.coolwarm,
        }

        return palette_options[self.settings.get("palette", "viridis")]

    def __init__(self):
        """Sets the initial context and initializes the Bokeh server."""
        self.settings = None
        self.app_context = {
            "data": None,
            "plots": {},
            "figures": {},
            "color_maps": {},
            "widgets": {},
        }

        self.init_bokeh_server()

    def __getattribute__(self, item):
        """Allow more convenient use of attributes from self.app_context. This is a bit strange.

        Args:
            item

        Returns:
            The resolved attribute if it is found in `self.app_context`.
        """
        try:
            return super().__getattribute__(item)
        except AttributeError:
            if item in self.app_context:
                return self.app_context[item]

    @abstractmethod
    def tool_handler(self, doc):
        """Hook for the application configuration and widget definition, without boilerplate."""
        pass

    def make_tool(
        self,
        arr: Union[xr.DataArray, str],
        notebook_url=None,
        notebook_handle=True,
        **kwargs
    ):
        """Starts the Bokeh application in accordance with the Bokeh app docs.

        Attempts to just guess the correct URL for Jupyter which is very error prone.
        """
        from bokeh.application import Application
        from bokeh.application.handlers import FunctionHandler
        from bokeh.io import show

        def generate_url(port):
            if port is None:
                return "localhost:8888"

            return "localhost:{}".format(port)

        if notebook_url is None:
            if "PORT" in SETTINGS:
                notebook_url = "localhost:{}".format(SETTINGS["PORT"])
            else:
                notebook_url = "localhost:8888"

        if isinstance(arr, str):
            arr = load_data(arr)
            if "cycle" in arr.dims and len(arr.dims) > 3:
                warnings.warn("Summing over cycle")
                arr = arr.sum("cycle", keep_attrs=True)

        if self.auto_zero_nans and {"kx", "ky", "kz", "kp"}.intersection(set(arr.dims)):
            # We need to copy and make sure to clear any nan values, because bokeh
            # does not send these over the wire for some reason
            arr = arr.copy()
            np.nan_to_num(arr.values, copy=False)

        # rebin any axes that have more than 800 pixels
        if self.auto_rebin and np.any(np.asarray(arr.shape) > self.rebin_size):
            reduction = {d: (s // self.rebin_size) + 1 for d, s in arr.S.dshape.items()}
            warnings.warn("Rebinning with {}".format(reduction))

            arr = rebin(arr, reduction=reduction)

            # TODO pass in a reference to the original copy of the array and make sure that
            # preparation tasks move over transparently

        self.arr = arr
        handler = FunctionHandler(self.tool_handler)
        app = Application(handler)
        show(app, notebook_url=notebook_url, notebook_handle=notebook_handle)

        return self.app_context


class SaveableTool(BokehInteractiveTool):
    def __init__(self, name=None):
        super().__init__()
        self.name = name
        self._last_save = None

    def make_tool(
        self,
        arr: Union[xr.DataArray, str],
        notebook_url=None,
        notebook_handle=True,
        **kwargs
    ):
        super().make_tool(
            arr, notebook_url=notebook_url, notebook_handle=notebook_handle, **kwargs
        )
        return self.app_context

    @property
    def filename(self):
        if self.name is None:
            return None

        return os.path.join(os.getcwd(), "tools", "tool-{}.json".format(self.name))

    @property
    def path(self):
        return None if self.filename is None else Path(self.filename)

    def load_app(self):
        if self.name is None:
            return

        if not self.path.exists():
            return {}

        self.path.parent.mkdir(exist_ok=True)
        with open(self.filename, "r") as f:
            self.deserialize(json.load(f))

    def save_app(self):
        if self.name is None:
            return

        data = self.serialize()
        if deep_equals(data, self._last_save):
            return

        self.path.parent.mkdir(exist_ok=True)
        with open(self.filename, "w") as f:
            self._last_save = data
            json.dump(data, f)

    def deserialize(self, json_data):
        pass

    def serialize(self):
        return {}



"""Very simple regression baselines."""

import pytorch_lightning as pl
from torch import nn, optim
import torch.nn.functional as F

__all__ = ["BaselineRegression", "LinearRegression"]


class LinearRegression(pl.LightningModule):
    """Linear regression, the simplest baseline."""

    input_dimensions = 200 * 200
    output_dimensions = 1

    def __init__(self):
        """Generate network components and use the mean squared error loss."""
        super().__init__()
        self.linear = nn.Linear(self.input_dimensions, self.output_dimensions)
        self.criterion = F.mse_loss

    def forward(self, x):
        """Calculate the model output for the minibatch `x`."""
        flat_x = x.view(x.size(0), -1)
        return self.linear(flat_x)

    def training_step(self, batch, batch_index):
        """Perform one training minibatch."""
        x, y = batch
        return self.criterion(self(x), y)

    def validation_step(self, batch, batch_index):
        """Perform one validation minibatch and record the validation loss."""
        x, y = batch
        loss = self.criterion(self(x), y)
        self.log("val_loss", loss)
        return loss

    def configure_optimizers(self):
        """Use standard optimizer settings."""
        return optim.Adam(self.parameters(), lr=3e-3)


class BaselineRegression(pl.LightningModule):
    """A simple three layer network providing a lowish capacity baseline."""

    input_dimensions = 200 * 200
    output_dimensions = 1

    def __init__(self):
        """Generate network components, matching data dimensions, and use MSE loss."""
        super().__init__()
        self.l1 = nn.Linear(self.input_dimensions, 256)
        self.l2 = nn.Linear(256, 128)
        self.l3 = nn.Linear(128, self.output_dimensions)
        self.criterion = F.mse_loss

    def forward(self, x):
        """Calculate the model output for the minibatch `x`."""
        flat_x = x.view(x.size(0), -1)
        h1 = F.relu(self.l1(flat_x))
        h2 = F.relu(self.l2(h1))
        output = self.l3(h2)
        return output

    def training_step(self, batch, batch_index):
        """Perform one training minibatch."""
        x, y = batch
        return self.criterion(self(x).squeeze(), y)

    def validation_step(self, batch, batch_index):
        """Perform one validation minibatch and record the validation loss."""
        x, y = batch
        loss = self.criterion(self(x).squeeze(), y)
        self.log("val_loss", loss)
        return loss

    def configure_optimizers(self):
        """Use standard optimizer settings."""
        return optim.Adam(self.parameters(), lr=3e-3)



"""Definitions of models involving Dirac points, graphene, graphite."""

from lmfit.models import update_param_vals
import lmfit as lf

from .x_model_mixin import XModelMixin
from .functional_forms import lorentzian

__all__ = [
    "DiracDispersionModel",
]


class DiracDispersionModel(XModelMixin):
    """Model for dirac_dispersion symmetric about the dirac point."""

    def dirac_dispersion(x, kd=1.6, amplitude_1=1, amplitude_2=1, center=0, sigma_1=1, sigma_2=1):
        """Model for dirac_dispersion symmetric about the dirac point.

        Fits lorentziants to (kd-center) and (kd+center)

        Args:
            x: value to evaluate fit at
            kd: Dirac point momentum
            amplitude_1: amplitude of Lorentzian at kd-center
            amplitude_2: amplitude of Lorentzian at kd+center
            center: center of Lorentzian
            sigma_1: FWHM of Lorentzian at kd-center
            sigma_2: FWHM of Lorentzian at kd+center

        Returns:
            An MDC model for a Dirac like dispersion around the cone.
        """
        return lorentzian(x, center=kd - center, amplitude=amplitude_1, gamma=sigma_1) + lorentzian(
            x, center=kd + center, amplitude=amplitude_2, gamma=sigma_2
        )

    def __init__(self, independent_vars=("x",), prefix="", missing="raise", name=None, **kwargs):
        """Defer to lmfit for initialization."""
        kwargs.update({"prefix": prefix, "missing": missing, "independent_vars": independent_vars})
        super().__init__(self.dirac_dispersion, **kwargs)

        self.set_param_hint("sigma_1", min=0.0)
        self.set_param_hint("sigma_2", min=0.0)

    def guess(self, data, x=None, **kwargs):
        """Placeholder for making better heuristic guesses here."""
        pars = self.make_params()

        # pars['%skd' % self.prefix].set(value=1.5)

        return update_param_vals(pars, self.prefix, **kwargs)

    __init__.doc = lf.models.COMMON_INIT_DOC
    guess.__doc__ = lf.models.COMMON_GUESS_DOC



"""Shows cut orientations for scans in Brillouin zones."""

import warnings

import numpy as np
import xarray as xr

from PyQt5 import QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvas
from matplotlib.figure import Figure

from arpes.utilities.conversion import convert_coordinates
from arpes.utilities.bz_spec import SURFACE_ZONE_DEFINITIONS
from arpes.utilities.image import imread_to_xarray
from arpes.utilities.qt import SimpleWindow, BasicHelpDialog
from arpes.utilities.qt.ui import horizontal, tabs, combo_box
from arpes.plotting.utils import imshow_arr
import arpes.xarray_extensions

from .CoordinateOffsetWidget import CoordinateOffsetWidget

__all__ = [
    "bz_tool",
]


class BZToolWindow(SimpleWindow):
    HELP_DIALOG_CLS = BasicHelpDialog


class BZTool:
    """Implements a Brillouin zone explorer showing the region of momentum probed by ARPES."""

    def __init__(self):
        self.settings = None
        self.context = {}

        self.content_layout = None
        self.main_layout = None
        self.views = {}
        self.reactive_views = []
        self.current_material = None
        self.cut_line = None

        self.canvas = None
        self.ax = None

    def configure_main_widget(self):
        self.canvas = FigureCanvas(Figure(figsize=(8, 8)))
        self.ax = self.canvas.figure.subplots()
        self.content_layout.addWidget(self.canvas, 0, 0)

    def on_change_material(self, value):
        self.current_material = SURFACE_ZONE_DEFINITIONS[value]

        if "bz_points" in self.current_material:
            bz_points = self.current_material["bz_points"]
            try:
                bz_points = bz_points()
            except TypeError:
                pass

            x, y = np.concatenate([bz_points, bz_points[:1]]).T
            self.ax.clear()
            self.ax.plot(x, y, linewidth=2)
            self.ax.set_aspect(1)
        else:
            assert "image" in self.current_material
            image = self.current_material["image"]
            image_waypoints = self.current_material["image_waypoints"]
            if not image_waypoints[0]:
                warnings.warn(
                    "Missing waypoints for material: {}".format(
                        self.current_material["name"]
                    )
                )
                image_waypoints = [
                    [0, 1, 0, 1],
                    [1, 0, 1, 0],
                ]

            self.image_waypoints = image_waypoints
            self.image_data = imread_to_xarray(image)

            # the way we are currently using waypoints assumes that the data has orthogonal
            # axes and they are not rotated. This might not be appropriate in all circumstances,
            # but it is how we are assuming things for now

            way_p, way_pp = image_waypoints
            way_p_x, way_p_y, way_p_kx, way_p_ky = way_p
            way_pp_x, way_pp_y, way_pp_kx, way_pp_ky = way_pp

            dx, dy, dkx, dky = (
                (way_pp_x - way_p_x),
                (way_pp_y - way_p_y),
                (way_pp_kx - way_p_kx),
                (way_pp_ky - way_p_ky),
            )

            self.image_data.coords["y"].values = (
                self.image_data.coords["y"].values - way_p_x
            ) * (dkx / dx) + way_p_kx
            self.image_data.coords["x"].values = (
                self.image_data.coords["x"].values - way_p_y
            ) * (-dky / dy) - way_p_ky

            self.ax.clear()
            self.image_data.values = self.image_data.values[::-1, :]
            imshow_arr(self.image_data, ax=self.ax, origin="lower")
            self.ax.scatter([0], [0], s=100, color="green")
            self.ax.set_aspect(1)

        self.ax.figure.canvas.draw()
        self.update_cut()

    @property
    def coordinates(self):
        return {k: np.pi * v.value() / 180 for k, v in self.coordinate_widgets.items()}

    def update_cut(self, *args):
        coords = {
            "phi": np.linspace(-15.0, 15.0, 51) * np.pi / 180,
            "hv": 5.93,  # FOR NOW
            "eV": 0,
        }

        coords_from_widgets = self.coordinates
        coords.update(
            {
                k: v
                for k, v in coords_from_widgets.items()
                if k
                not in {
                    "phi",
                }
            }
        )

        cut = xr.Dataset(
            data_vars={},
            coords=coords,
            attrs={
                "work_function": self.current_material.get("work_function", 4.2) or 4.2,
                "inner_potential": self.current_material.get("inner_potential", 10)
                or 10,
                "phi_offset": coords_from_widgets["phi"],
            },
        )

        kcut = convert_coordinates(cut)

        try:
            self.ax.lines.remove(self.cut_line[0])
        except Exception:
            pass

        self.cut_line = self.ax.plot(
            kcut.kx.values,
            kcut.ky.values,
            self.current_material.get("cut_color", "green"),
            linewidth=3,
        )
        self.ax.figure.canvas.draw()

    def construct_coordinate_info_tab(self):
        needed_coordinates = ["phi", "psi", "alpha", "theta", "beta", "chi"]
        inner_items = [
            CoordinateOffsetWidget(coordinate_name=coordinate, root=self)
            for coordinate in needed_coordinates
        ]
        self.coordinate_widgets = dict(zip(needed_coordinates, inner_items))

        for widget in inner_items:
            pass
        return horizontal(*inner_items)

    def construct_sample_info_tab(self):
        material_choice = combo_box(
            sorted(SURFACE_ZONE_DEFINITIONS.keys()),
            name="Material Specification",
        )
        self.material_choice_widget = material_choice
        self.material_choice_widget.currentTextChanged.connect(self.on_change_material)
        # self.photon_energy_widget =
        inner_items = [material_choice]
        return horizontal(*inner_items)

    def construct_detector_info_tab(self):
        inner_items = []
        return horizontal(*inner_items)

    def construct_general_settings_tab(self):
        inner_items = []
        return horizontal(*inner_items)

    def add_coordinate_control_widgets(self):
        self.coordinate_info_tab = self.construct_coordinate_info_tab()
        self.sample_info_tab = self.construct_sample_info_tab()
        self.detector_info_tab = self.construct_detector_info_tab()
        self.general_settings_tab = self.construct_general_settings_tab()

        self.tabs = tabs(
            ["Coordinates", self.coordinate_info_tab],
            ["Sample Info", self.sample_info_tab],
            ["Detector Type and Settings", self.detector_info_tab],
            # ['Beamline Info', self.beamline_info],
            ["General Settings", self.general_settings_tab],
        )

        self.tabs.setFixedHeight(250)
        self.main_layout.addLayout(self.content_layout, 0, 0)
        self.main_layout.addWidget(self.tabs, 1, 0)

    def start(self):
        app = QtWidgets.QApplication([])

        win = BZToolWindow()
        win.resize(1100 - 250, 1100)
        win.setWindowTitle("Brillouin Zone Tool")
        cw = QtWidgets.QWidget()
        win.setCentralWidget(cw)
        self.window = win
        self.window.tool = self

        self.content_layout = QtWidgets.QGridLayout()
        self.main_layout = QtWidgets.QGridLayout()

        cw.setLayout(self.main_layout)

        self.configure_main_widget()
        self.add_coordinate_control_widgets()

        self.material_choice_widget.setCurrentText("Td-WTe2")

        self.window.show()

        QtWidgets.QApplication.instance().exec()


def bz_tool():
    """Starts the Brillouin zone exploration tool."""
    tool = BZTool()
    tool.start()



"""Application infrastructure for apps/tools which browse a data volume."""

import sys
from math import floor
from typing import Any, TYPE_CHECKING
import weakref
import numpy as np
import xarray as xr
import pyqtgraph as pg
from PyQt5.QtWidgets import QApplication, QWidget
from collections import defaultdict

from arpes.utilities.qt.ui import CursorRegion
from arpes.typing import DataType
from arpes.settings import SETTINGS
from .data_array_image_view import DataArrayImageView, DataArrayPlot
from .utils import PlotOrientation, ReactivePlotRecord


if TYPE_CHECKING:
    from .windows import SimpleWindow

__all__ = ["SimpleApp"]


class SimpleApp(QApplication):
    """
    Has all of the layout information and business logic for an interactive data
    browsing utility using PyQt5.
    """

    WINDOW_CLS = None
    WINDOW_SIZE = (4, 4)
    TITLE = "Untitled Tool"

    _data: xr.Dataset = None
    spectrum: xr.DataArray = None

    def __init__(self):
        """Only interesting thing on init is to make a copy of the user settings."""
        super().__init__(sys.argv)
        self._ninety_eight_percentile = None
        self.settings = None
        self._window = None
        self._layout = None

        self.context = {}

        self.views = {}
        self.reactive_views: list[ReactivePlotRecord] = []
        self.registered_cursors: dict[str, list[CursorRegion]] = defaultdict(list)

        self.settings = SETTINGS.copy()

    def copy_to_clipboard(self, value: Any) -> None:
        """Attempts to copy the value to the clipboard, or else prints."""
        try:
            import pyperclip
            import pprint

            pyperclip.copy(pprint.pformat(value))
        except ImportError:
            pass
        finally:
            import pprint

            print(pprint.pformat(value))

    @property
    def data(self) -> DataType:
        """
        Read data from the cached attribute.

        This is a property as opposed to a plain attribute
        in order to facilitate rendering datasets with several
        data_vars.
        """
        return self._data

    @data.setter
    def data(self, new_data: DataType):
        self._data = new_data
        self.spectrum = (
            new_data.S.spectrum if isinstance(new_data, xr.Dataset) else new_data
        )

    def close(self):
        """
        Graceful shutdown. Tell each view to close and drop references so GC happens.
        """
        for v in self.views.values():
            v.close()

        self.views = {}
        self.reactive_views = []

    @property
    def ninety_eight_percentile(self):
        """
        Calculates the 98 percentile of data so colorscale is not outlier dependent.
        """
        if self._ninety_eight_percentile is not None:
            return self._ninety_eight_percentile

        self._ninety_eight_percentile = np.percentile(self.spectrum.values, 98)
        return self._ninety_eight_percentile

    def print(self, *args, **kwargs):
        """Forwards printing to the application so it ends up in Jupyter."""
        self.window.window_print(*args, **kwargs)

    @staticmethod
    def build_pg_cmap(colormap):
        """Converts a matplotlib colormap to one suitable for pyqtgraph.

        pyqtgraph uses its own colormap format but for consistency and aesthetic
        reasons we want to use the ones from matplotlib. This will sample the colors
        from the colormap and convert it into an array suitable for pyqtgraph.
        """
        sampling_array = np.linspace(0, 1, 5)
        sampled_colormap = colormap(sampling_array) * 255

        return pg.ColorMap(
            pos=np.linspace(0, 1, len(sampled_colormap)), color=sampled_colormap
        )

    def set_colormap(self, colormap):
        """Finds all `DataArrayImageView` instances and sets their color palette."""
        import matplotlib.cm

        if isinstance(colormap, str):
            colormap = matplotlib.cm.get_cmap(colormap)

        cmap = self.build_pg_cmap(colormap)
        for view in self.views.values():
            if isinstance(view, DataArrayImageView):
                view.setColorMap(cmap)

    def generate_marginal_for(
        self,
        dimensions: tuple[str],
        row: int,
        column: int,
        name: str = None,
        orientation: PlotOrientation = PlotOrientation.Horizontal,
        cursors: bool = False,
        layout=None,
    ):
        """
        Generates a marginal plot for this applications's data after selecting along
        `dimensions`. This is used to generate the many different views of a volume in
        the browsable tools.
        """
        layout = self._layout if layout is None else layout

        remaining_dims = [dim for dim in self.spectrum.dims if dim not in dimensions]

        if len(remaining_dims) == 1:
            widget = DataArrayPlot(
                name=name, root=weakref.ref(self), orientation=orientation
            )
            self.views[name] = widget

            if orientation == PlotOrientation.Horizontal:
                widget.setMaximumHeight(200)
            else:
                widget.setMaximumWidth(200)

            if cursors:
                remaining_dim = remaining_dims[0]
                cursor = CursorRegion(
                    orientation=(
                        CursorRegion.Horizontal
                        if orientation == PlotOrientation.Vertical
                        else CursorRegion.Vertical
                    ),
                    movable=True,
                    bounds=[
                        0,
                        self.spectrum.sizes[remaining_dim],
                    ],
                )
                widget.addItem(cursor, ignoreBounds=False)
                self.connect_cursor(remaining_dim, cursor)
        else:
            assert len(remaining_dims) == 2
            widget = DataArrayImageView(name=name, root=weakref.ref(self))
            widget.view.setAspectLocked(False)
            self.views[name] = widget

            # TODO: set the bottom level intelligently
            widget.setHistogramRange(0, self.ninety_eight_percentile)
            widget.setLevels(0.05, 0.95)

            if cursors:
                for dim, orientation in zip(
                    remaining_dims, [CursorRegion.Vertical, CursorRegion.Horizontal]
                ):
                    cursor = CursorRegion(
                        orientation=orientation,
                        movable=True,
                        bounds=[0, self.spectrum.sizes[dim]],
                    )
                    widget.addItem(cursor, ignoreBounds=True)
                    self.connect_cursor(dim, cursor)

        self.reactive_views.append(
            ReactivePlotRecord(dims=dimensions, view=widget, orientation=orientation)
        )
        layout.addWidget(widget, row, column)
        return widget

    def connect_cursor(self, dimension: str, the_line: CursorRegion) -> None:
        """Connect a cursor to a line control.

        without weak references we get a circular dependency here
        because `the_line` is owned by a child of `self` but we are
        providing self to a closure which is retained by `the_line`.
        """
        self.registered_cursors[dimension].append(the_line)
        owner = weakref.ref(self)

        def connected_cursor(line: CursorRegion):
            new_cursor = owner().context["cursor"].copy()
            new_cursor[dimension] = line.getRegion()
            owner().update_cursor_position(new_cursor)

        the_line.sigRegionChanged.connect(connected_cursor)

    def reconnect_cursor(self, dimension: str, the_line: CursorRegion) -> None:
        """
        Reconnect a cursor to a line control. User must remove the line from
        registered_cursors before calling this function or there will be a duplicate.
        """
        the_line.sigRegionChanged.disconnect()
        self.connect_cursor(dimension, the_line)

    def update_cursor_position(
        self, new_cursor: dict[str, tuple[float, float]], **kwargs
    ) -> None:
        """
        Sets the current cursor positions. If the cursor has changed, this also
        updates all reactive views.

        Args:
            new_cursor: The new cursor positions.
            kwargs: Additional arguments to pass to `update_reactive_views`.
        """
        old_cursor = self.context["cursor"].copy()
        changed_dimensions = set(
            dim for dim in new_cursor if new_cursor[dim] != old_cursor[dim]
        )
        if not changed_dimensions:
            return
        self.context["cursor"] = new_cursor

        def indices_to_values(indices: float, dim: str):
            coord_values = self.spectrum.coords[dim].values

            # TODO: interpolate for irregularly spaced coords
            def index_to_value(index):
                return coord_values[0] + index * (coord_values[1] - coord_values[0])

            return tuple(index_to_value(index) for index in indices)

        self.context["value_cursor"] = {
            dim: indices_to_values(indices, dim) for dim, indices in new_cursor.items()
        }
        cursor_text = " ".join(
            f"{dim}:({values[0]:.3g}, {values[1]:.3g})"
            for dim, values in self.context["value_cursor"].items()
        )
        self.window.statusBar().showMessage(cursor_text)

        self.update_reactive_views(
            changed_dimensions=changed_dimensions,
            **kwargs,
        )

    def update_reactive_views(
        self,
        coord_slices: dict[str, slice] = None,
        changed_dimensions: set[str] = None,
        force=False,
        keep_levels=True,
        reactive_views=None,
    ):
        """
        Updates all reactive views based on the specified slices.

        Args:
            coord_slices: The slices to apply to the data.
            changed_dimensions: The dimensions that have changed.
            force: If true, will update all views regardless of whether they have
                changed.
            keep_levels: If true, will keep the existing colormap levels.
        """
        if coord_slices is None:
            cursor: dict[str, tuple[float, float]] = self.context["cursor"]
            coord_slices = {dim: slice(*position) for dim, position in cursor.items()}
        changed_dimensions = set() if changed_dimensions is None else changed_dimensions

        def safe_slice(unsafe_slice: slice, dim: str):
            start, stop = floor(unsafe_slice.start), floor(unsafe_slice.stop)
            dim_length = self.spectrum.sizes[dim]
            start, stop = np.clip(start, 0, dim_length), np.clip(stop, 0, dim_length)
            if start == stop:
                stop = start + 1
            start, stop = np.clip(start, 0, dim_length), np.clip(stop, 0, dim_length)
            if start == stop:
                start = stop - 1
            return slice(start, stop)

        safe_slices = {
            dim: safe_slice(unsafe_slice, dim)
            for dim, unsafe_slice in coord_slices.items()
        }
        reactive_views = (
            self.reactive_views if reactive_views is None else reactive_views
        )
        for plot_record in reactive_views:
            if set(plot_record.dims).intersection(changed_dimensions) or force:
                reactive_slices = {dim: safe_slices[dim] for dim in plot_record.dims}
                view = plot_record.view

                if isinstance(view, DataArrayImageView):
                    image_data = view.root.spectrum.isel(**reactive_slices)
                    if reactive_slices:
                        image_data = image_data.mean(list(reactive_slices.keys()))
                    view.setImage(image_data, keep_levels=keep_levels)

                elif isinstance(view, DataArrayPlot):
                    plot_data = view.root.spectrum.isel(**reactive_slices)
                    if reactive_slices:
                        plot_data = plot_data.mean(list(reactive_slices.keys()))

                    cursors = [
                        plot_item
                        for plot_item in view.getPlotItem().items
                        if isinstance(plot_item, CursorRegion)
                    ]
                    view.clear()
                    for cursor in cursors:
                        view.addItem(cursor)

                    view.plot(plot_data)

    def before_show(self):
        """Lifecycle hook."""
        pass

    def after_show(self):
        """Lifecycle hook."""
        pass

    def layout(self):
        """
        Hook for defining the application layout. This needs to be provided by
        subclasses.
        """
        raise NotImplementedError

    @property
    def window(self) -> "SimpleWindow":
        """Gets the window instance on the current application."""
        return self._window

    def start(self):
        """Starts the Qt application, configures the window, and begins Qt execution."""

        from arpes.utilities.qt import qt_info

        qt_info.init_from_app(self)

        self._window = self.WINDOW_CLS()
        self.window.resize(*qt_info.inches_to_px(self.WINDOW_SIZE))
        self.window.setWindowTitle(self.TITLE)

        self.central_widget = QWidget()
        self._layout = self.layout()
        self.central_widget.setLayout(self._layout)
        self.window.setCentralWidget(self.central_widget)
        self.window.app = weakref.ref(self)

        self.before_show()
        self.window.show()
        self.after_show()

        qt_info.apply_settings_to_app(self)

        self.exec()



"""This module contains utilities for generating momentum grids from angle-space data.

This process consists of:
    1. Determining the momentum axes which are necessary for a dataset based on which coordinate axes it has
    2. Determining the range over the output axes which is required for the data
    3. Determining an appropriate resolution or binning in the output grid
"""

import enum
import itertools
from typing import List, Optional

__all__ = [
    "is_dimension_unconvertible",
    "AxisType",
    "determine_axis_type",
    "determine_momentum_axes_from_measurement_axes",
]


def is_dimension_unconvertible(dimension_name: str) -> bool:
    """Determines whether a dimension does not participate in the momentum conversion.

    Many axes, like temperature, are just along for the ride and so we can just pass through
    the conversion.

    Args:
        dimension_name (str): [description]

    Returns:
        bool: [description]
    """
    if dimension_name in [
        "eV",
        "delay",
        "cycle",
        "temp",
        "temperature",
        "x",
        "y",
        "optics_insertion",
    ]:
        return True

    if "volt" in dimension_name:
        return True

    return False


class AxisType(str, enum.Enum):
    """Models whether a given dimension is angle-like or momentum-like."""

    Angle = "angle"
    Momentum = "k"


def determine_axis_type(coordinate_names: List[str], permissive: bool = True) -> AxisType:
    """Determines whether the input axes are better described as angle axes or momentum axes.

    Args:
        coordinate_names: The names of the coordinates
        permissive: Whether additional coordinates should be tossed out before checking

    Returns:
        What kind of axes they are.
    """
    coordinate_names = tuple(sorted(coordinate_names))
    mapping = {
        ("beta", "phi"): "angle",
        ("chi", "phi"): "angle",
        ("phi", "psi"): "angle",
        ("phi", "theta"): "angle",
        ("kx", "ky"): "kp",
        ("kx", "kz"): "k",
        ("ky", "kz"): "k",
        ("kx", "ky", "kz"): "k",
    }

    all_allowable = set(itertools.chain(*mapping.keys()))
    fixed_coordinate_names = tuple(t for t in coordinate_names if t in all_allowable)

    if fixed_coordinate_names != coordinate_names and not permissive:
        raise ValueError(
            f"""Received some coordinates {coordinate_names} which are
                not compatible with angle/k determination."""
        )

    return mapping[coordinate_names]


def determine_momentum_axes_from_measurement_axes(axis_names: List[str]) -> Optional[List[str]]:
    """Associates the appropriate set of momentum dimensions given the angular dimensions."""
    axis_names = tuple(sorted(axis_names))

    return {
        ("phi",): ["kp"],
        ("beta", "phi"): ["kx", "ky"],
        ("phi", "theta"): ["kx", "ky"],
        ("phi", "psi"): ["kx", "ky"],
        ("hv", "phi"): ["kp", "kz"],
        ("beta", "hv", "phi"): ["kx", "ky", "kz"],
        ("hv", "phi", "theta"): ["kx", "ky", "kz"],
        ("hv", "phi", "psi"): ["kx", "ky", "kz"],
    }.get(axis_names)



"""Contains utilities for performing statistical operations in spectra and DataArrays."""

import xarray as xr
from arpes.provenance import update_provenance
from arpes.utilities import lift_dataarray_to_generic

__all__ = ("mean_and_deviation",)


@update_provenance("Calculate mean and standard deviation for observation axis")
@lift_dataarray_to_generic
def mean_and_deviation(data: xr.DataArray, axis=None, name=None):
    """Calculates the mean and standard deviation of a DataArray along an axis.

    The reduced axis corresponds to individual observations of a tensor/array valued quantity.
    This axis can be passed or inferred from a set of standard observation-like axes.

    New data variables are created with names `{name}` and `{name}_std`.
    If a name is not attached to the DataArray, it should be provided.

    Args:
        data: The input data.
        axis: The name of the dimension which we should perform the reduction along.
        name: The name of the variable which should be reduced. By default, uses `data.name`.

    Returns:
        A dataset with variables corresponding to the mean and standard error of each
        relevant variable in the input DataArray.
    """
    preferred_axes = ["bootstrap", "cycle", "idx"]

    name = data.name if data.name is not None else name
    assert name is not None

    if axis is None:
        for pref_axis in preferred_axes:
            if pref_axis in data.dims:
                axis = pref_axis
                break

    assert axis in data.dims
    return xr.Dataset(
        data_vars={name: data.mean(axis), name + "_std": data.std(axis)}, attrs=data.attrs
    )



"""Tools for analyzing moirés and data on moiré heterostructures in particular.

All of the moirés discussed here are on hexagonal crystal systems.
"""

import matplotlib.pyplot as plt

from arpes.utilities.bz import hex_cell_2d
from arpes.plotting.bz import bz_plot, Translation, Rotation

import numpy as np
from scipy.spatial.distance import pdist

__all__ = [
    "mod_points_to_lattice",
    "generate_other_lattice_points",
    "unique_points",
    "plot_simple_moire_unit_cell",
    "calc_commensurate_moire_cell",
    "angle_between_vectors",
]


def higher_order_commensurability(lattice_constant_ratio, order=2, angle_range=None):
    if angle_range is None:
        angle_range = (0, 5)

    bz_corner_a, bz_corner_b = np.asarray([1, 0]), np.asarray([0.5, np.sqrt(3) / 2])

    return bz_corner_a, bz_corner_b


def mod_points_to_lattice(pts, a, b):
    """Projects points to lattice equivalent ones in the first primitive cell."""
    rmat = np.asarray([[0, -1], [1, 0]])
    ra, rb = rmat @ a, rmat @ b
    ra, rb = ra / (np.sqrt(3) / 2), rb / (np.sqrt(3) / 2)

    return pts + np.outer(np.ceil(pts @ rb), a) + np.outer(np.ceil(pts @ -ra), b)


def generate_other_lattice_points(a, b, ratio, order=1, angle=0):
    """Generates (a, b, angle) superlattice points."""
    ratio = max(np.abs(ratio), 1 / np.abs(ratio))
    cosa, sina = np.cos(angle), np.sin(angle)
    rmat = np.asarray([[cosa, -sina], [sina, cosa]])
    a, b = rmat @ (ratio * a), rmat @ (ratio * b)

    ias = np.arange(-order, order + 1)
    pts = (a[None, None, :] * ias[None, :, None]) + (b[None, None, :] * ias[:, None, None])
    pts = pts.reshape(len(ias) ** 2, 2)

    # not quite correct, since we need the manhattan distance
    # mask = np.linalg.norm(pts, axis=1) <= order * ratio

    ds = np.stack(
        [
            (np.outer(ias[None, :], (ias[None, :] * 0 + 1))),
            (np.outer(ias[None, :] * 0 + 1, (ias[None, :]))),
        ],
        axis=-1,
    ).reshape(len(ias) ** 2, 2)

    dabs = np.abs(np.sum(ds, axis=1))
    dist = np.max(np.abs(ds), axis=1)
    sign = np.sign(ds)
    sign = sign[:, 0] == sign[:, 1]
    dist[sign] = dabs[sign]

    return pts[dist <= order]


def unique_points(pts):
    """Makes a collection of points unique by removing duplicates."""
    return np.vstack([np.array(u) for u in set([tuple(p) for p in pts])])


def generate_segments(grouped_points, a, b):
    moded = mod_points_to_lattice(grouped_points, a, b)
    g1d = np.diff(np.sum(grouped_points, axis=1))
    m1d = np.diff(np.sum(moded, axis=1))

    low_index = 0
    for split_index in np.nonzero(np.abs((m1d - g1d)) > 1e-11)[0]:
        yield moded[low_index : split_index + 1]
        low_index = split_index + 1

    yield moded[low_index:]


def minimum_distance(pts, a, b):
    moded = np.stack([mod_points_to_lattice(x, a, b) for x in pts], axis=1)
    return np.min(np.stack([pdist(x) for x in moded], axis=-1), axis=0)


def calculate_bz_vertices_from_direct_cell(cell):
    from ase.dft.bz import bz_vertices

    if len(cell) > 2:
        assert all(abs(cell[2][0:2]) < 1e-6) and all(abs(cell.T[2][0:2]) < 1e-6)
    else:
        cell = [list(c) + [0] for c in cell] + [[0, 0, 1]]

    icell = np.linalg.inv(cell).T
    try:
        bz1 = bz_vertices(icell[:3, :3], dim=2)
    except TypeError:
        bz1 = bz_vertices(icell[:3, :3])

    return bz1


def angle_between_vectors(a, b) -> float:
    """Calculates the angle between two vectors using the law of cosines."""
    return np.arccos(np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b)))


def calc_commensurate_moire_cell(underlayer_a, overlayer_a, relative_angle=0, swap_angle=False):
    """Calculates nearly commensurate moire unit cells for two hexagonal lattices."""
    from ase.dft.kpoints import get_special_points
    from ase.dft.bz import bz_vertices

    underlayer_direct = hex_cell_2d(a=underlayer_a)
    overlayer_direct = hex_cell_2d(a=overlayer_a)

    underlayer_direct = [list(c) + [0] for c in underlayer_direct] + [[0, 0, 1]]
    overlayer_direct = [list(c) + [0] for c in overlayer_direct] + [[0, 0, 1]]

    underlayer_icell = np.linalg.inv(underlayer_direct).T
    overlayer_icell = np.linalg.inv(overlayer_direct).T

    underlayer_k = np.dot(underlayer_icell.T, get_special_points(underlayer_direct)["K"])
    overlayer_k = Rotation.from_rotvec([0, 0, relative_angle]).apply(
        np.dot(overlayer_icell.T, get_special_points(overlayer_direct)["K"])
    )

    moire_k = underlayer_k - overlayer_k
    moire_a = underlayer_a * (np.linalg.norm(underlayer_k) / np.linalg.norm(moire_k))
    moire_angle = angle_between_vectors(underlayer_k, moire_k)

    if swap_angle:
        moire_angle = -moire_angle

    moire_cell = hex_cell_2d(moire_a)
    moire_cell = [list(c) + [0] for c in moire_cell] + [[0, 0, 1]]
    moire_cell = Rotation.from_rotvec([0, 0, moire_angle]).apply(moire_cell)
    moire_icell = np.linalg.inv(moire_cell).T

    moire_bz_points = bz_vertices(moire_icell)
    moire_bz_points = moire_bz_points[[len(p[0]) for p in moire_bz_points].index(6)][0]

    return {
        "k_points": (underlayer_k, overlayer_k, moire_k),
        "moire_a": moire_a,
        "moire_k": moire_k,
        "moire_cell": moire_cell,
        "moire_icell": moire_icell,
        "moire_bz_points": moire_bz_points,
        "moire_bz_angle": moire_angle,
    }


def plot_simple_moire_unit_cell(
    underlayer_a, overlayer_a, relative_angle, ax=None, offset=True, swap_angle=False
):
    """Plots a digram of a moiré unit cell."""
    if ax is None:
        _, ax = plt.subplots()

    bz_plot(
        cell=hex_cell_2d(a=underlayer_a),
        linewidth=1,
        ax=ax,
        paths=[],
        hide_ax=False,
        set_equal_aspect=False,
    )
    bz_plot(
        cell=hex_cell_2d(a=overlayer_a),
        linewidth=1,
        ax=ax,
        paths=[],
        transformations=[Rotation.from_rotvec([0, 0, relative_angle])],
        hide_ax=False,
        set_equal_aspect=False,
    )

    moire_info = calc_commensurate_moire_cell(
        underlayer_a, overlayer_a, relative_angle, swap_angle=swap_angle
    )
    moire_k = moire_info["moire_k"]

    if offset:
        k_offset = Rotation.from_rotvec([0, 0, 120 * np.pi / 180]).apply(moire_k)
    else:
        k_offset = 0

    bz_plot(
        cell=hex_cell_2d(a=moire_info["moire_a"]),
        linewidth=1,
        ax=ax,
        paths=[],
        transformations=[
            Rotation.from_rotvec([0, 0, -moire_info["moire_bz_angle"]]),
            Translation(moire_info["k_points"][0] + k_offset),
        ],
        hide_ax=True,
        set_equal_aspect=True,
    )



"""Utilities for comparing collections and some specialty collection types."""

from collections.abc import Mapping

import numpy as np

from typing import Any

__all__ = (
    "deep_equals",
    "deep_update",
    "MappableDict",
)


class MappableDict(dict):
    """Like dict except that +, -, *, / are cascaded to values."""

    def __add__(self, other):
        """Applies `+` onto values."""
        if set(self.keys()) != set(other.keys()):
            raise ValueError("You can only add two MappableDicts with the same keys.")

        return MappableDict({k: self.get(k) + other.get(k) for k in self.keys()})

    def __sub__(self, other):
        """Applies `-` onto values."""
        if set(self.keys()) != set(other.keys()):
            raise ValueError(
                "You can only subtract two MappableDicts with the same keys."
            )

        return MappableDict({k: self.get(k) - other.get(k) for k in self.keys()})

    def __mul__(self, other):
        """Applies `*` onto values."""
        if set(self.keys()) != set(other.keys()):
            raise ValueError(
                "You can only multiply two MappableDicts with the same keys."
            )

        return MappableDict({k: self.get(k) * other.get(k) for k in self.keys()})

    def __truediv__(self, other):
        """Applies `/` onto values."""
        if set(self.keys()) != set(other.keys()):
            raise ValueError(
                "You can only divide two MappableDicts with the same keys."
            )

        return MappableDict({k: self.get(k) / other.get(k) for k in self.keys()})

    def __floordiv__(self, other):
        """Applies `//` onto values."""
        if set(self.keys()) != set(other.keys()):
            raise ValueError(
                "You can only divide (//) two MappableDicts with the same keys."
            )

        return MappableDict({k: self.get(k) // other.get(k) for k in self.keys()})

    def __neg__(self):
        """Applies unary negation onto values."""
        return MappableDict({k: -self.get(k) for k in self.keys()})


def deep_update(destination: Any, source: Any) -> dict[str, Any]:
    """Doesn't clobber keys further down trees like doing a shallow update would.

    Instead recurse down from the root and update as appropriate.

    Args:
        destination
        source

    Returns:
        The destination item
    """
    for k, v in source.items():
        if isinstance(v, Mapping):
            destination[k] = deep_update(destination.get(k, {}), v)
        else:
            destination[k] = v

    return destination


def deep_equals(a: Any, b: Any) -> bool:
    """An equality check that looks into common collection types."""
    if not isinstance(b, type(a)):
        print(b, a)
        return False

    if isinstance(
        a,
        (
            int,
            str,
            float,
            np.float32,
            np.int32,
            np.float64,
            np.int64,
        ),
    ):
        return a == b

    if a is None:
        return b is None

    if not isinstance(
        a,
        (
            dict,
            list,
            tuple,
            set,
        ),
    ):
        raise TypeError(
            f"Only dict, list, tuple, and set are supported by deep_equals, not {type(a)}"
        )

    if isinstance(a, set):
        for item in a:
            if item not in b:
                return False

        for item in b:
            if item not in a:
                return False

        return True

    if isinstance(a, (list, tuple)):
        if len(a) != len(b):
            return False

        for i in range(len(a)):
            item_a, item_b = a[i], b[i]

            if not deep_equals(item_a, item_b):
                return False

        return True

    if isinstance(a, dict):
        if not set(a.keys()) == set(b.keys()):
            return False

        for k in a.keys():
            item_a, item_b = a[k], b[k]

            if not deep_equals(item_a, item_b):
                return False

        return True



"""Some functional and UI functional programming utilities."""
import functools
import time
from collections import defaultdict

from numpy import ndarray

import xarray as xr
from arpes.typing import DataType
from typing import Any, Callable, Dict, Iterator, Optional, Tuple

__all__ = [
    "Debounce",
    "lift_dataarray_to_generic",
    "iter_leaves",
    "group_by",
    "cycle",
]


def cycle(sequence):
    """Infinitely cycles a sequence."""
    while True:
        for s in sequence:
            yield s


def group_by(grouping, sequence):
    """Permits partitining a sequence into sets of items, for instance by taking two at a time."""
    if isinstance(grouping, int):
        base_seq = [False] * grouping
        base_seq[-1] = True

        grouping = cycle(base_seq)

    groups = []
    current_group = []
    for elem in sequence:
        current_group.append(elem)

        if (callable(grouping) and grouping(elem)) or next(grouping):
            groups.append(current_group)
            current_group = []

    if len(current_group):
        groups.append(current_group)

    return groups


def collect_leaves(tree: Dict[str, Any], is_leaf: Optional[Any] = None) -> Dict:
    """Produces a flat representation of the leaves.

    Leaves with the same key are collected into a list in the order of appearance,
    but this depends on the dictionary iteration order.

    Example:
    collect_leaves({'a': 1, 'b': 2, 'c': {'a': 3, 'b': 4}}) -> {'a': [1, 3], 'b': [2, 4]}

    Args:
        tree: The nested dictionary structured tree
        is_leaf: A condition to determine whether the current node is a leaf

    Returns:
        A dictionary with the leaves and their direct parent key.
    """

    def reducer(dd: Dict, item: Tuple[str, ndarray]) -> Dict:
        dd[item[0]].append(item[1])
        return dd

    return functools.reduce(reducer, iter_leaves(tree, is_leaf), defaultdict(list))


def iter_leaves(
    tree: Dict[str, Any], is_leaf: Optional[Callable] = None
) -> Iterator[Tuple[str, ndarray]]:
    """Iterates across the leaves of a nested dictionary.

    Whether a particular piece
    of data counts as a leaf is controlled by the predicate `is_leaf`. By default,
    all nested dictionaries are considered not leaves, i.e. an item is a leaf if and
    only if it is not a dictionary.

    Iterated items are returned as key value pairs.

    As an example, you can easily flatten a nested structure with
    `dict(leaves(data))`
    """
    if is_leaf is None:
        is_leaf = lambda x: not isinstance(x, dict)

    for k, v in tree.items():
        if is_leaf(v):
            yield k, v
        else:
            for item in iter_leaves(v):
                yield item


def lift_dataarray_to_generic(f):
    """A functorial decorator that lifts functions to operate over xarray types.

    (xr.DataArray, *args, **kwargs) -> xr.DataArray

    to one with signature

    A = typing.Union[xr.DataArray, xr.Dataset]
    (A, *args, **kwargs) -> A

    i.e. one that will operate either over xr.DataArrays or xr.Datasets.
    """

    @functools.wraps(f)
    def func_wrapper(data: DataType, *args, **kwargs):
        if isinstance(data, xr.DataArray):
            return f(data, *args, **kwargs)
        else:
            assert isinstance(data, xr.Dataset)
            new_vars = {datavar: f(data[datavar], *args, **kwargs) for datavar in data.data_vars}

            for var_name, var in new_vars.items():
                if isinstance(var, xr.DataArray) and var.name is None:
                    var.name = var_name

            merged = xr.merge(new_vars.values())
            return merged.assign_attrs(data.attrs)

    return func_wrapper


class Debounce:
    """Wraps a function so that it can only be called periodically.

    Very useful for preventing expensive recomputation of some UI state when a user
    is performing a continuous action like a mouse pan or scroll or manipulating a
    slider.
    """

    def __init__(self, period):
        """Sets up the internal state for debounce tracking."""
        self.period = period  # never call the wrapped function more often than this (in seconds)
        self.count = 0  # how many times have we successfully called the function
        self.count_rejected = 0  # how many times have we rejected the call
        self.last = None  # the last time it was called

    def reset(self):
        """Force a reset of the timer, aka the next call will always work."""
        self.last = None

    def __call__(self, f):
        """The wrapper call which defers execution if the function was actually called recently."""

        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            now = time.time()
            willcall = False
            if self.last is not None:
                # amount of time since last call
                delta = now - self.last
                if delta >= self.period:
                    willcall = True
                else:
                    willcall = False
            else:
                willcall = True  # function has never been called before

            if willcall:
                # set these first incase we throw an exception
                self.last = now  # don't use time.time()
                self.count += 1
                f(*args, **kwargs)  # call wrapped function
            else:
                self.count_rejected += 1

        return wrapped



"""Utilities related to plotting Brillouin zones and data onto them."""
# pylint: disable=import-error

import itertools
import warnings

import matplotlib.cm
import matplotlib.pyplot as plt
import numpy as np

from typing import Union, List

from matplotlib.patches import FancyArrowPatch
from mpl_toolkits.mplot3d import proj3d
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from scipy.spatial.transform import Rotation

from arpes.analysis.mask import apply_mask_to_coords
from arpes.plotting.utils import path_for_plot
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.bz import build_2dbz_poly, process_kpath, hex_cell_2d
from arpes.utilities.bz_spec import A_GRAPHENE, A_WS2, A_WSe2
from arpes.utilities.geometry import polyhedron_intersect_plane

__all__ = (
    "annotate_special_paths",
    "bz2d_plot",
    "bz3d_plot",
    "bz_plot",
    "plot_data_to_bz",
    "plot_data_to_bz2d",
    "plot_data_to_bz3d",
    "plot_plane_to_bz",
    "bz2d_segments",
    "overplot_standard",
)

overplot_library = {
    "graphene": lambda: {"cell": hex_cell_2d(A_GRAPHENE)},
    "ws2": lambda: {"cell": hex_cell_2d(A_WS2)},
    "wwe2": lambda: {"cell": hex_cell_2d(A_WSe2)},
}


def segments_standard(name="graphene", rotate=0):
    name = name.lower()
    specification = overplot_library[name]()
    transformations = []
    if rotate:
        transformations = [Rotation.from_rotvec([0, 0, rotate])]

    return bz2d_segments(specification["cell"], transformations)


def overplot_standard(name="graphene", repeat=None, rotate=0):
    """A higher order function to plot a Brillouin zone over a plot."""
    specification = overplot_library[name]()
    transformations = []

    if rotate:
        transformations = [Rotation.from_rotvec([0, 0, rotate])]

    def overplot_the_bz(ax):
        bz_plot(
            cell=specification["cell"],
            linewidth=2,
            ax=ax,
            paths=[],
            repeat=repeat,
            set_equal_aspect=False,
            hide_ax=False,
            transformations=transformations,
            zorder=5,
            linestyle="-",
        )

    return overplot_the_bz


class Translation:
    """Base translation class, meant to provide some extension over rotations.

    Rotations are available from `scipy.spatial.transform.Rotation`.
    """

    translation_vector = None

    def __init__(self, translation_vector):
        print(translation_vector)
        self.translation_vector = np.asarray(translation_vector)

    def apply(self, vectors, inverse=False):
        """Applies the translation to a set of vectors.

        If this transform is D-dimensional (for D=2,3) and is applied to a different
        dimensional set of vectors, a ValueError will be thrown due to the dimension
        mismatch.

        An inverse flag is available in order to apply the inverse coordinate transform.
        Up to numerical accuracy,

        ```
        self.apply(self.apply(vectors), inverse=True) == vectors
        ```

        Args:
            vectors: array_like with shape (2 or 3,) or (N, 2 or 3)
            inverse: Applies the inverse coordinate transform instead
        """
        vectors = np.asarray(vectors)

        if vectors.ndim > 2 or vectors.shape[-1] not in {2, 3}:
            raise ValueError(
                "Expected a 2D or 3D vector (2 or 3,) of list of vectors (N, 2 or 3,), instead "
                "recevied: {}".format(vectors.shape)
            )

        single_vector = False
        if vectors.ndim == 1:
            single_vector = True
            vectors = vectors[None, :]  # expand dims

        if inverse:
            result = vectors - self.translation_vector
        else:
            result = vectors + self.translation_vector

        return result if not single_vector else result[0]


Transformation = Union[Translation, Rotation]


def apply_transformations(
    points: np.ndarray, transformations: List[Transformation] = None, inverse=False
) -> np.ndarray:
    """Applies a series of transformations to a sequence of vectors or a single vector.

    Args:
        points
        translations
        inverse

    Returns:
        The collection of transformed points.
    """
    if transformations is None:
        transformations = []

    for transformation in transformations:
        points = transformation.apply(points, inverse=inverse)

    return points


def plot_plane_to_bz(cell, plane, ax, special_points=None, facecolor="red"):
    """Plots a 2D cut plane onto a Brillouin zone."""
    from ase.dft.bz import bz_vertices

    if isinstance(plane, str):
        plane_points = process_kpath(plane, cell, special_points=special_points)[0]
    else:
        plane_points = plane

    d1, d2 = plane_points[1] - plane_points[0], plane_points[2] - plane_points[0]

    faces = [p[0] for p in bz_vertices(np.linalg.inv(cell).T)]
    pts = polyhedron_intersect_plane(faces, np.cross(d1, d2), plane_points[0])

    collection = Poly3DCollection([pts])
    collection.set_facecolor(facecolor)
    ax.add_collection3d(collection, zs="z")


def plot_data_to_bz(data: DataType, cell, **kwargs):
    """A dimension agnostic tool used to plot ARPES data onto a Brillouin zone."""
    if len(data) == 3:
        return plot_data_to_bz3d(data, cell, **kwargs)

    return plot_data_to_bz2d(data, cell, **kwargs)


def plot_data_to_bz2d(
    data: DataType,
    cell,
    rotate=None,
    shift=None,
    scale=None,
    ax=None,
    mask=True,
    out=None,
    bz_number=None,
    **kwargs
):
    """Plots data onto a 2D Brillouin zone."""
    data = normalize_to_spectrum(data)

    assert "You must k-space convert data before plotting to BZs" and data.S.is_kspace

    if bz_number is None:
        bz_number = (0, 0)

    fig = None
    if ax is None:
        fig, ax = plt.subplots(figsize=(9, 9))
        bz2d_plot(cell, paths="all", ax=ax)

    if len(cell) == 2:
        cell = [list(c) + [0] for c in cell] + [[0, 0, 1]]

    icell = np.linalg.inv(cell).T

    # Prep coordinates and mask
    raveled = data.G.meshgrid(as_dataset=True)
    dims = data.dims
    if rotate is not None:
        c, s = np.cos(rotate), np.sin(rotate)
        rotation = np.array([(c, -s), (s, c)])

        raveled = raveled.G.transform_coords(dims, rotation)

    if scale is not None:
        raveled = raveled.G.scale_coords(dims, scale)

    if shift is not None:
        raveled = raveled.G.shift_coords(dims, shift)

    copied = data.values.copy()

    if mask:
        built_mask = apply_mask_to_coords(raveled, build_2dbz_poly(cell=cell), dims)
        copied[built_mask.T] = np.nan

    cmap = kwargs.get("cmap", matplotlib.cm.Blues)
    if isinstance(cmap, str):
        cmap = matplotlib.cm.get_cmap(cmap)

    cmap.set_bad((1, 1, 1, 0))

    delta_x = np.dot(np.array(bz_number), icell[:2, 0])
    delta_y = np.dot(np.array(bz_number), icell[:2, 1])

    ax.pcolormesh(
        raveled.data_vars[dims[0]].values + delta_x,
        raveled.data_vars[dims[1]].values + delta_y,
        copied.T,
        cmap=cmap,
    )

    if out is not None:
        plt.savefig(path_for_plot(out), dpi=400)
        return path_for_plot(out)

    return fig, ax


def plot_data_to_bz3d(data: DataType, cell, **kwargs):
    """Plots ARPES data onto a 3D Brillouin zone."""
    raise NotImplementedError("plot_data_to_bz3d is not implemented yet.")


def bz_plot(cell, *args, **kwargs):
    """Dimension generic BZ plot which uses the cell dimension to delegate."""
    if len(cell) > 2:
        return bz3d_plot(cell, *args, **kwargs)

    return bz2d_plot(cell, *args, **kwargs)


def bz3d_plot(
    cell,
    vectors=False,
    paths=None,
    points=None,
    ax=None,
    elev=None,
    scale=1,
    repeat=None,
    transformations=None,
    hide_ax=True,
    **kwargs
):
    """For now this is lifted from ase.dft.bz.bz3d_plot with some modifications.

    All copyright and licensing terms for this and bz2d_plot are those of the current release of ASE
    (Atomic Simulation Environment).
    """
    try:
        from ase.dft.bz import bz_vertices  # dynamic because we do not require ase
    except ImportError:
        warnings.warn(
            "You will need to install ASE (Atomic Simulation Environment) to use this feature."
        )
        raise ImportError("You will need to install ASE before using Brillouin Zone plotting")

    class Arrow3D(FancyArrowPatch):
        def __init__(self, xs, ys, zs, *args, **kwargs):
            FancyArrowPatch.__init__(self, (0, 0), (0, 0), *args, **kwargs)
            self._verts3d = xs, ys, zs

        def draw(self, renderer):
            xs3d, ys3d, zs3d = self._verts3d
            xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, renderer.M)
            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
            FancyArrowPatch.draw(self, renderer)

    icell = np.linalg.inv(cell).T
    kpoints = points

    if isinstance(paths, str):
        from ase.dft.kpoints import (
            get_special_points,
            special_paths,
            parse_path_string,
            crystal_structure_from_cell,
        )

        special_points = get_special_points(cell)
        structure = crystal_structure_from_cell(cell)

        path_string = special_paths[structure] if paths == "all" else paths
        paths = []
        for names in parse_path_string(path_string):
            points = []
            for name in names:
                points.append(np.dot(icell.T, special_points[name]))
                paths.append((names, points))

    if ax is None:
        fig = plt.figure(figsize=(5, 5))
        ax = fig.gca(projection="3d")

    azim = np.pi / 5
    elev = elev or np.pi / 6
    x = np.sin(azim)
    y = np.cos(azim)
    view = [x * np.cos(elev), y * np.cos(elev), np.sin(elev)]

    bz1 = bz_vertices(icell)

    maxp = 0.0

    if repeat is None:
        repeat = (
            1,
            1,
            1,
        )

    dx, dy, dz = icell[0], icell[1], icell[2]
    rep_x, rep_y, rep_z = repeat

    if isinstance(rep_x, int):
        rep_x = (0, rep_x)
    if isinstance(rep_y, int):
        rep_y = (0, rep_y)
    if isinstance(rep_z, int):
        rep_z = (0, rep_z)

    c = kwargs.pop("c", "k")
    c = kwargs.pop("color", c)

    for ix, iy, iz in itertools.product(range(*rep_x), range(*rep_y), range(*rep_z)):
        delta = dx * ix + dy * iy + dz * iz

        for points, normal in bz1:
            color = c

            if np.dot(normal, view) < 0:
                ls = ":"
            else:
                ls = "-"

            cosines = np.dot(icell, normal) / np.linalg.norm(normal) / np.linalg.norm(icell, axis=1)
            for idx, cosine in enumerate(cosines):
                if np.abs(np.abs(cosine) - 1) < 1e-6 and False:  # debugging this
                    tup = [rep_x, rep_y, rep_z][idx]
                    current = [ix, iy, iz][idx]

                    if cosine < 0:
                        current = current - 1

                    if tup[0] < current + 1 < tup[1]:
                        color = (1, 1, 1, 0)

                    if current + 1 != tup[1] and current != tup[0]:
                        ls = ":"
                        color = "blue"

            x, y, z = np.concatenate([points, points[:1]]).T
            x, y, z = x + delta[0], y + delta[1], z + delta[2]

            ax.plot(x, y, z, c=color, ls=ls, **kwargs)
            maxp = max(maxp, points.max())

    if vectors:
        ax.add_artist(
            Arrow3D(
                [0, icell[0, 0]],
                [0, icell[0, 1]],
                [0, icell[0, 2]],
                mutation_scale=20,
                lw=1,
                arrowstyle="-|>",
                color="k",
            )
        )
        ax.add_artist(
            Arrow3D(
                [0, icell[1, 0]],
                [0, icell[1, 1]],
                [0, icell[1, 2]],
                mutation_scale=20,
                lw=1,
                arrowstyle="-|>",
                color="k",
            )
        )
        ax.add_artist(
            Arrow3D(
                [0, icell[2, 0]],
                [0, icell[2, 1]],
                [0, icell[2, 2]],
                mutation_scale=20,
                lw=1,
                arrowstyle="-|>",
                color="k",
            )
        )
        maxp = max(maxp, 0.6 * icell.max())

    if paths is not None:
        for names, points in paths:
            x, y, z = np.array(points).T
            ax.plot(x, y, z, c="r", ls="-")

            for name, point in zip(names, points):
                x, y, z = point
                if name == "G":
                    name = "\\Gamma"
                elif len(name) > 1:
                    name = name[0] + "_" + name[1]
                ax.text(x, y, z, "$" + name + "$", ha="center", va="bottom", color="r")

    if kpoints is not None:
        for p in kpoints:
            ax.scatter(p[0], p[1], p[2], c="b")

    if hide_ax:
        ax.set_axis_off()
        ax.autoscale_view(tight=True)

    s = maxp / 0.5 * 0.45 * scale
    ax.set_xlim(-s, s)
    ax.set_ylim(-s, s)
    ax.set_zlim(-s, s)
    ax.set_aspect("equal")

    ax.view_init(azim=azim / np.pi * 180, elev=elev / np.pi * 180)


def annotate_special_paths(
    ax,
    paths,
    cell=None,
    transformations=None,
    offset=None,
    special_points=None,
    labels=None,
    **kwargs
):
    """Annotates user indicated paths in k-space by plotting lines (or points) over the BZ."""
    if paths == "":
        raise ValueError("Must provide a proper path.")

    if isinstance(paths, (list, str)):
        if isinstance(paths, str):
            if labels is None:
                labels = paths

            paths = process_kpath(paths, cell, special_points=special_points)

            if not isinstance(labels[0], list):
                labels = [labels]

            labels = [list(l) for l in labels]
            paths = list(zip(labels, paths))

    fontsize = kwargs.pop("fontsize", 14)

    if offset is None:
        offset = dict()

    two_d = True
    try:
        ax.get_zlim()
        two_d = False
    except AttributeError:
        pass

    for names, points in paths:
        x, y, z = np.array(points).T

        if two_d:
            ax.plot(x, y, c="r", ls="-", **kwargs)
        else:
            ax.plot(x, y, z, c="r", ls="-", **kwargs)

        for name, point in zip(names, points):
            x, y, z = point
            display_name = name
            if display_name == "G":
                display_name = "\\Gamma"
            elif len(name) > 1:
                name = name[0] + "_" + name[1]

            off = offset.get(name, 0)
            try:
                if two_d:
                    x, y = x + off[0], y + off[1]
                else:
                    x, y, z = x + off[0], y + off[1], z + off[2]

            except TypeError:
                if two_d:
                    x, y = x + off, y + off
                else:
                    x, y, z = x + off, y + off, z + off

            if two_d:
                ax.text(
                    x,
                    y,
                    "$" + display_name + "$",
                    ha="center",
                    va="bottom",
                    color="r",
                    fontsize=fontsize,
                )
            else:
                ax.text(
                    x,
                    y,
                    z,
                    "$" + display_name + "$",
                    ha="center",
                    va="bottom",
                    color="r",
                    fontsize=fontsize,
                )


def bz2d_segments(cell, transformations=None):
    """Calculates the line segments corresponding to a 2D BZ."""
    segments_x = []
    segments_y = []

    for points, normal in twocell_to_bz1(cell)[0]:
        points = apply_transformations(points, transformations)
        x, y, z = np.concatenate([points, points[:1]]).T
        segments_x.append(x)
        segments_y.append(y)

    return segments_x, segments_y


def twocell_to_bz1(cell):
    from ase.dft.bz import bz_vertices

    # 2d in x-y plane
    if len(cell) > 2:
        assert all(abs(cell[2][0:2]) < 1e-6) and all(abs(cell.T[2][0:2]) < 1e-6)
    else:
        cell = [list(c) + [0] for c in cell] + [[0, 0, 1]]
    icell = np.linalg.inv(cell).T
    try:
        bz1 = bz_vertices(icell[:3, :3], dim=2)
    except TypeError:
        bz1 = bz_vertices(icell[:3, :3])
    return bz1, icell, cell


def bz2d_plot(
    cell,
    vectors=False,
    paths=None,
    points=None,
    repeat=None,
    ax=None,
    transformations=None,
    offset=None,
    hide_ax=True,
    set_equal_aspect=True,
    **kwargs
):
    """This piece of code modified from ase.ase.dft.bz.py:bz2d_plot and follows copyright and license for ASE.

    Plots a Brillouin zone corresponding to a given unit cell
    """
    kpoints = points
    bz1, icell, cell = twocell_to_bz1(cell)
    if ax is None:
        ax = plt.axes()

    if isinstance(paths, str):
        from ase.dft.kpoints import (
            get_special_points,
            special_paths,
            parse_path_string,
            crystal_structure_from_cell,
        )

        special_points = get_special_points(cell)
        structure = crystal_structure_from_cell(cell)

        path_string = special_paths[structure] if paths == "all" else paths
        paths = []
        for names in parse_path_string(path_string):
            points = []
            for name in names:
                points.append(np.dot(icell.T, special_points[name]))
                paths.append((names, points))

    maxp = 0.0
    c = kwargs.pop("c", "k")
    c = kwargs.pop("color", c)
    ls = kwargs.pop("ls", kwargs.pop("linestyle", "-"))

    for points, normal in bz1:
        points = apply_transformations(points, transformations)
        x, y, z = np.concatenate([points, points[:1]]).T

        ax.plot(x, y, c=c, ls=ls, **kwargs)
        maxp = max(maxp, points.max())

    if repeat is not None:
        dx, dy = icell[0], icell[1]

        rep_x, rep_y = repeat
        if isinstance(rep_x, int):
            rep_x = (0, rep_x)
        if isinstance(rep_y, int):
            rep_y = (0, rep_y)

        for ix, iy in itertools.product(range(*rep_x), range(*rep_y)):
            delta = dx * ix + dy * iy

            for points, normal in bz1:
                x, y, z = np.concatenate([points, points[:1]]).T
                x, y = x + delta[0], y + delta[1]
                x, y, z = apply_transformations(np.asarray([x, y, z]).T, transformations).T

                c = kwargs.pop("c", "k")
                c = kwargs.pop("color", c)
                ax.plot(x, y, c=c, ls=ls, **kwargs)
                maxp = max(maxp, points.max())

    if vectors:
        ax.arrow(
            0,
            0,
            icell[0, 0],
            icell[0, 1],
            lw=1,
            color="k",
            length_includes_head=True,
            head_width=0.03,
            head_length=0.05,
        )
        ax.arrow(
            0,
            0,
            icell[1, 0],
            icell[1, 1],
            lw=1,
            color="k",
            length_includes_head=True,
            head_width=0.03,
            head_length=0.05,
        )

    if paths is not None:
        annotate_special_paths(ax, paths, offset=offset, transformations=transformations)

    if kpoints is not None:
        for p in kpoints:
            ax.scatter(p[0], p[1], c="b")

    if hide_ax:
        ax.set_axis_off()
        ax.autoscale_view(tight=True)

    if set_equal_aspect:
        ax.set_aspect("equal")



"""Provides broadcasted and process parallel curve fitting for PyARPES.

The core of this module is `broadcast_model` which is a serious workhorse in PyARPES for
analyses based on curve fitting. This allows simple multidimensional curve fitting by
iterative fitting across one or many axes. Currently basic strategies are implemented,
but in the future we would like to provide:

1. Passing xr.DataArray values to parameter guesses and bounds, which can be interpolated/selected
   to allow changing conditions throughout the curve fitting session.
2. A strategy allowing retries with initial guess taken from the previous fit. This is similar
   to some adaptive curve fitting routines that have been proposed in the literature.
"""

import os
import dill
from packaging import version

import lmfit
import numpy as np
from tqdm import tqdm_notebook

import arpes.fits.fit_models
from typing import Callable, Union, Tuple, Any, Dict, List

import xarray as xr
from arpes.provenance import update_provenance
from arpes.typing import DataType
from arpes.utilities import normalize_to_spectrum
from . import mp_fits

__all__ = ("broadcast_model", "result_to_hints")


TypeIterable = Union[List[type], Tuple[type]]

XARRAY_REQUIRES_VALUES_WRAPPING = version.parse(xr.__version__) > version.parse(
    "0.10.0"
)


def wrap_for_xarray_values_unpacking(item):
    """This is a shim for https://github.com/pydata/xarray/issues/2097."""
    if XARRAY_REQUIRES_VALUES_WRAPPING:
        return np.array(item, dtype=object)

    return item


def result_to_hints(
    m: lmfit.model.ModelResult, defaults=None
) -> Dict[str, Dict[str, Any]]:
    """Turns an `lmfit.model.ModelResult` into a dictionary with initial guesses.

    Args:
        m: The model result to extract parameters from
        defaults: Returned if `m` is None, useful for cell re-evaluation in Jupyter

    Returns:
        A dict containing parameter specifications in key-value rathan than `lmfit.Parameter`
        format, as you might pass as `params=` to PyARPES fitting code.
    """
    if m is None:
        return defaults
    return {k: {"value": m.params[k].value} for k in m.params}


def parse_model(model):
    """Takes a model string and turns it into a tokenized version.

    1. ModelClass -> ModelClass
    2. [ModelClass] -> [ModelClass]
    3. str -> [<ModelClass, operator as string>]

    i.e.

    A + (B + C) * D -> [A, '(', B, '+', C, ')', '*', D]

    Args:
        model: The model specification

    Returns:
        A tokenized specification of the model suitable for passing to the curve
        fitting routine.
    """
    if not isinstance(model, str):
        return model

    pad_all = ["+", "-", "*", "/", "(", ")"]

    for pad in pad_all:
        model = model.replace(pad, " {} ".format(pad))

    special = set(pad_all)

    def read_token(token):
        if token in special:
            return token
        try:
            token = float(token)
            return token
        except ValueError:
            try:
                return arpes.fits.fit_models.__dict__[token]
            except KeyError:
                raise ValueError("Could not find model: {}".format(token))

    return [read_token(token) for token in model.split()]


@update_provenance("Broadcast a curve fit along several dimensions")
def broadcast_model(
    model_cls: Union[type, TypeIterable],
    data: DataType,
    broadcast_dims,
    params=None,
    progress=True,
    weights=None,
    safe=False,
    prefixes=None,
    window=None,
    parallelize=None,
):
    """Perform a fit across a number of dimensions.

    Allows composite models as well as models defined and compiled through strings.

    Args:
        model_cls: The model specification
        data: The data to curve fit
        broadcast_dims: Which dimensions of the input should be iterated across as opposed
          to fit across
        params: Parameter hints, consisting of plain values or arrays for interpolation
        progress: Whether to show a progress bar
        weights: Weights to apply when curve fitting. Should have the same shape as the input data
        safe: Whether to mask out nan values
        window: A specification of cuts/windows to apply to each curve fit
        parallelize: Whether to parallelize curve fits, defaults to True if unspecified and more
          than 20 fits were requested

    Returns:
        An `xr.Dataset` containing the curve fitting results. These are data vars:

        - "results": Containing an `xr.DataArray` of the `lmfit.model.ModelResult` instances
        - "residual": The residual array, with the same shape as the input
        - "data": The original data used for fitting
        - "norm_residual": The residual array normalized by the data, i.e. the fractional error
    """
    if params is None:
        params = {}

    if isinstance(broadcast_dims, str):
        broadcast_dims = [broadcast_dims]

    data = normalize_to_spectrum(data)

    cs = {}
    for dim in broadcast_dims:
        cs[dim] = data.coords[dim]

    other_axes = set(data.dims).difference(set(broadcast_dims))
    template = data.sum(list(other_axes))
    template.values = np.ndarray(template.shape, dtype=object)
    n_fits = np.prod(np.array(list(template.S.dshape.values())))

    if parallelize is None:
        parallelize = n_fits > 20

    residual = data.copy(deep=True)
    residual.values = np.zeros(residual.shape)

    model = parse_model(model_cls)

    wrap_progress = lambda x, *_, **__: x
    if progress:
        wrap_progress = tqdm_notebook

    serialize = parallelize
    fitter = mp_fits.MPWorker(
        data=data,
        uncompiled_model=model,
        prefixes=prefixes,
        params=params,
        safe=safe,
        serialize=serialize,
        weights=weights,
        window=window,
    )

    if parallelize:
        # print("Running on multiprocessing pool... this may take a while the first time.")
        from .hot_pool import hot_pool

        pool = hot_pool.pool
        exe_results = list(
            wrap_progress(
                pool.imap(fitter, template.G.iter_coords()),
                total=n_fits,
                desc="Fitting on pool...",
            )
        )
    else:
        exe_results = []
        for _, cut_coords in wrap_progress(
            template.G.enumerate_iter_coords(), desc="Fitting", total=n_fits
        ):
            exe_results.append(fitter(cut_coords))

    if serialize:
        # print("Deserializing...")

        def unwrap(result_data):
            # using the lmfit deserialization and serialization seems slower than double pickling with dill
            # result = lmfit.model.ModelResult(compiled_model, compiled_model.make_params())
            # return result.loads(result_data)
            return dill.loads(result_data)

        exe_results = [(unwrap(res), residual, cs) for res, residual, cs in exe_results]
        # print("Finished deserializing")

    for fit_result, fit_residual, coords in exe_results:
        template.loc[coords] = wrap_for_xarray_values_unpacking(fit_result)
        residual.loc[coords] = fit_residual

    return xr.Dataset(
        {
            "results": template,
            "data": data,
            "residual": residual,
            "norm_residual": residual / data,
        },
        residual.coords,
    )



"""Provides plotting formatters for different kinds of data and targets."""
from typing import Dict, Any

__all__ = [
    "SpectrumFormatter",
    "FloatTitleFormatter",
]


class SpectrumFormatter:
    """Knows how to plot an ARPES spectrum onto an interpretation plot."""

    def show(self, data, ax=None):
        """Just imshow the data for now with no other decoration."""
        spectrum, row = data
        ax.imshow(spectrum, origin="lower")


class FloatTitleFormatter:
    """Plots a floating point target as a title annotation onto a plot for its parent item."""

    context: Dict[str, Any] = None
    title_formatter: str = r"{label}={data:.3f}"

    @property
    def computed_context(self) -> Dict[str, Any]:
        """Annotate whether this is a ground truth or predicted value."""
        return {"label": "True" if self.context.get("is_ground_truth", False) else "Pred"}

    def show(self, data, ax=None):
        """Sets the title for the parent data axis to be the formatted float value."""
        title = ax.get_title()
        context = {
            **self.computed_context,
            "data": data,
        }

        ax.set_title(
            "{title}; {addendum}".format(
                title=title, addendum=self.title_formatter.format(**context)
            )
        )



"""Collect imports from categorized submodules."""
from .x_model_mixin import *
from .wrapped import *
from .backgrounds import *
from .peaks import *
from .fermi_edge import *
from .dirac import *
from .two_dimensional import *
from .misc import *
from .decay import *



"""Provides utilities used internally by `arpes.analysis.band_analysis`."""
from collections import namedtuple

import numpy as np

ParamType = namedtuple("ParamType", ["value", "stderr"])


def param_getter(param_name, safe=True):
    """Constructs a function to extract a parameter value by name.

    Useful to extract data from inside an array of `lmfit.ModelResult` instances.

    Args:
        param_name: Parameter name to retrieve. If you performed a
          composite model fit, make sure to include the prefix.
        safe: Guards against NaN values. This is typically desirable but
          sometimes it is advantageous make sure to include the prefix.
          to have NaNs fail an analysis quickly.

    Returns:
        A function which fetches the fitted value for this named parameter.
    """
    if safe:
        safe_param = ParamType(value=np.nan, stderr=np.nan)

        def getter(x):
            try:
                return x.params.get(param_name, safe_param).value
            except:
                return np.nan

        return getter

    return lambda x: x.params[param_name].value


def param_stderr_getter(param_name, safe=True):
    """Constructs a function to extract a parameter value by name.

    Useful to extract data from inside an array of `lmfit.ModelResult` instances.

    Args:
        param_name: Parameter name to retrieve. If you performed a
          composite model fit, make sure to include the prefix.
        safe: Guards against NaN values. This is typically desirable but
          sometimes it is advantageous make sure to include the prefix.
          to have NaNs fail an analysis quickly.

    Returns:
        A function which fetches the standard error for this named parameter.
    """
    if safe:
        safe_param = ParamType(value=np.nan, stderr=np.nan)

        def getter(x):
            try:
                return x.params.get(param_name, safe_param).stderr
            except:
                return np.nan

        return getter

    return lambda x: x.params[param_name].stderr



"""Provides some simple analysis tools in Qt format. Useful for selecting regions and points."""

import pyqtgraph as pg
import numpy as np
import xarray as xr
from scipy import interpolate

from PyQt5 import QtWidgets, QtCore

from arpes import analysis
from arpes.utilities import normalize_to_spectrum
from arpes.utilities.conversion import DetectorCalibration
from arpes.utilities.qt import qt_info, SimpleApp, SimpleWindow, BasicHelpDialog
from arpes.typing import DataType
from arpes.utilities.qt.ui import KeyBinding

__all__ = (
    "path_tool",
    "mask_tool",
    "bkg_tool",
    "det_window_tool",
)


qt_info.setup_pyqtgraph()


class CoreToolWindow(SimpleWindow):
    HELP_DIALOG_CLS = BasicHelpDialog

    def compile_key_bindings(self):
        return super().compile_key_bindings() + [
            KeyBinding("Transpose - Roll Axis", [QtCore.Qt.Key_T], self.transpose_roll),
            KeyBinding(
                "Transpose - Swap Front Axes", [QtCore.Qt.Key_Y], self.transpose_swap
            ),
        ]

    def transpose_roll(self, event):
        self.app.transpose_to_front(-1)

    def transpose_swap(self, event):
        self.app.transpose_to_front(1)


class CoreTool(SimpleApp):
    WINDOW_SIZE = (
        5,
        6.5,
    )
    WINDOW_CLS = CoreToolWindow
    TITLE = ""
    ROI_CLOSED = False
    SUMMED = True

    def __init__(self):
        self.data = None
        self.roi = None
        self.main_layout = QtWidgets.QGridLayout()
        self.content_layout = QtWidgets.QGridLayout()

        super().__init__()

    def layout(self):
        return self.main_layout

    def set_data(self, data: DataType):
        self.data = normalize_to_spectrum(data)

    def transpose_to_front(self, dim):
        if not isinstance(dim, str):
            dim = self.data.dims[dim]

        order = list(self.data.dims)
        order.remove(dim)
        order = [dim] + order

        reindex_order = [self.data.dims.index(t) for t in order]
        self.data = self.data.transpose(*order)
        self.update_data()
        self.roi_changed(self.roi)

    def configure_image_widgets(self):
        if len(self.data.dims) == 3:
            self.generate_marginal_for(
                (0,), 0, 0, "xy", cursors=False, layout=self.content_layout
            )
            self.generate_marginal_for(
                (0,), 1, 0, "P", cursors=False, layout=self.content_layout
            )
        else:
            self.generate_marginal_for(
                (), 0, 0, "xy", cursors=False, layout=self.content_layout
            )

            if self.SUMMED:
                self.generate_marginal_for(
                    (0,), 1, 0, "P", cursors=False, layout=self.content_layout
                )
            else:
                self.generate_marginal_for(
                    (), 1, 0, "P", cursors=False, layout=self.content_layout
                )

        self.attach_roi()
        self.main_layout.addLayout(self.content_layout, 0, 0)

    def attach_roi(self):
        self.roi = pg.PolyLineROI([[0, 0], [50, 50]], closed=self.ROI_CLOSED)
        self.views["xy"].view.addItem(self.roi)
        self.roi.sigRegionChanged.connect(self.roi_changed)

    def compute_path_from_roi(self, roi):
        offset = roi.pos()
        points = roi.getState()["points"]
        x, y = [p.x() + offset.x() for p in points], [
            p.y() + offset.y() for p in points
        ]

        nx, ny = self.data.dims[-2], self.data.dims[-1]
        cx, cy = self.data.coords[nx], self.data.coords[ny]
        x = interpolate.interp1d(
            np.arange(len(cx)), cx.values, fill_value="extrapolate"
        )(x)
        y = interpolate.interp1d(
            np.arange(len(cy)), cy.values, fill_value="extrapolate"
        )(y)

        points = []
        for xi, yi in zip(x, y):
            points.append(dict([[nx, xi], [ny, yi]]))

        return points

    @property
    def path(self):
        return self.compute_path_from_roi(self.roi)

    def roi_changed(self, _):
        try:
            self.path_changed(self.path)
        except:
            pass

    def path_changed(self, path):
        raise NotImplementedError

    def add_controls(self):
        pass

    def update_data(self):
        if len(self.data.dims) == 3:
            self.views["xy"].setImage(
                self.data.S.nan_to_num(),
                xvals=self.data.coords[self.data.dims[0]].values,
            )
            self.views["P"].setImage(self.data.mean(self.data.dims[0]))
        else:
            self.views["xy"].setImage(self.data.S.nan_to_num())
            if self.SUMMED:
                self.views["P"].plot(
                    self.data.isel(**dict([[self.data.dims[0], 0]])) * 0
                )
            else:
                self.views["P"].setImage(self.data)

    def before_show(self):
        self.configure_image_widgets()
        self.add_controls()
        self.update_data()


class PathTool(CoreTool):
    TITLE = "Path-Tool"

    def path_changed(self, path):
        selected_data = self.data.S.along(path)
        if len(selected_data.dims) == 2:
            self.views["P"].setImage(selected_data.data.transpose())
        else:
            self.views["P"].clear()
            self.views["P"].plot(selected_data.data)


class DetectorWindowTool(CoreTool):
    TITLE = "Detector-Window"
    ROI_CLOSED = False
    alt_roi = None

    def attach_roi(self):
        spectrum = normalize_to_spectrum(self.data).S.sum_other(["eV", "phi"])
        spacing = int(len(spectrum.eV) / 10)
        take_eVs = [i * spacing for i in range(10)]
        if take_eVs[-1] != len(spectrum.eV) - 1:
            take_eVs += [len(spectrum.eV) - 1]

        left_ext = spectrum.X.first_exceeding(
            "phi", 0.25, relative=True, reverse=True, as_index=True
        )
        right_ext = spectrum.X.first_exceeding(
            "phi", 0.25, relative=True, as_index=True
        )

        xl, xr = take_eVs, take_eVs
        yl, yr = left_ext[take_eVs], right_ext[take_eVs]

        if spectrum.dims.index("eV") != 0:
            xl, yl = yl, xl
            xr, yr = yr, xr

        self.roi = pg.PolyLineROI(list(zip(xl, yl)), closed=self.ROI_CLOSED)
        self.alt_roi = pg.PolyLineROI(list(zip(xr, yr)), closed=self.ROI_CLOSED)
        self.views["xy"].view.addItem(self.roi)
        self.views["xy"].view.addItem(self.alt_roi)

    @property
    def alt_path(self):
        return self.compute_path_from_roi(self.alt_roi)

    def path_changed(self, path):
        pass

    @property
    def calibration(self):
        return DetectorCalibration(left=self.alt_path, right=self.path)


class MaskTool(CoreTool):
    TITLE = "Mask-Tool"
    ROI_CLOSED = True
    SUMMED = False

    @property
    def mask(self):
        path = self.path
        dims = self.data.dims[-2:]
        return {
            "dims": dims,
            "polys": [[[p[d] for d in dims] for p in path]],
        }

    def path_changed(self, _):
        mask = self.mask

        main_data = self.data
        if len(main_data.dims) > 2:
            main_data = main_data.isel(
                **dict([[main_data.dims[0], self.views["xy"].currentIndex]])
            )

        if len(mask["polys"][0]) > 2:
            self.views["P"].setImage(
                analysis.apply_mask(main_data, mask, replace=0, invert=True)
            )


class BackgroundTool(CoreTool):
    TITLE = "Background-Tool"
    ROI_CLOSED = False
    SUMMED = False

    @property
    def mode(self):
        return "slice"

    def path_changed(self, path):
        dims = self.data.dims[-2:]
        slices = {}
        for dim in dims:
            coordinates = [p[dim] for p in path]
            slices[dim] = slice(min(coordinates), max(coordinates))

        main_data = self.data
        if len(main_data.dims) > 2:
            main_data = main_data.isel(
                **dict([[main_data.dims[0], self.views["xy"].currentIndex]])
            )

        bkg = 0
        if self.mode == "slice":
            bkg = main_data.sel(
                **{k: v for k, v in slices.items() if k == dims[0]}
            ).mean(dims[0])

        self.views["P"].setImage(main_data - bkg)

    def add_controls(self):
        pass


def wrap(cls):
    def tool_function(data):
        tool = cls()
        tool.set_data(data)
        tool.start()
        return tool

    return tool_function


path_tool = wrap(PathTool)
mask_tool = wrap(MaskTool)
bkg_tool = wrap(BackgroundTool)
det_window_tool = wrap(DetectorWindowTool)



"""Facilitates saving intermediate data using a portable binary format for numpy."""
import json
from pathlib import Path

import numpy as np

__all__ = [
    "to_portable_bin",
    "from_portable_bin",
]

DTYPES = {
    "float64": np.float64,
    "float32": np.float32,
    "int64": np.int64,
    "int32": np.int32,
}


def from_portable_bin(path: Path) -> np.ndarray:
    """Reads data from a relatively portable binary format.

    A "portable" binary file is a directory containing
    a binary file and a very small json file which contains

    1. dtype
    2. shape
    3. offset

    We do this instead of using pickling in order to ensure that the
    data formats are portable.
    """
    with open(str(path / "portability.json"), "r") as f:
        portability = json.load(f)

    dtype = DTYPES[portability.pop("dtype")]
    shape = portability["shape"]

    arr = np.fromfile(str(path / "arr.bin"), dtype=dtype)
    arr = arr.reshape(shape)
    return arr


def to_portable_bin(arr: np.ndarray, path: Path) -> None:
    """Converts data to a relatively portable binary format.

    See also `read_portable_bin`.

    Writes array as a binary format with an associated json description
    of necessary portability info.
    """
    path.mkdir(parents=True, exist_ok=True)
    json_path, arr_path = path / "portability.json", path / "arr.bin"
    assert not (json_path.exists() or arr_path.exists())

    with open(str(json_path), "w") as f:
        json.dump(
            {
                "dtype": arr.dtype.name,
                "shape": arr.shape,
            },
            f,
        )

    arr.tofile(str(arr_path.resolve()))
